"""Pronunciation dictionaries for use in alignment and transcription"""

from __future__ import annotations

import math
import os
import subprocess
import sys
from collections import Counter, defaultdict
from typing import TYPE_CHECKING, Any, Collection, Optional

if TYPE_CHECKING:
    from montreal_forced_aligner.dictionary.multispeaker import MultispeakerDictionaryMixin
    from montreal_forced_aligner.abc import (
        ReversedMappingType,
    )

from montreal_forced_aligner.dictionary.mixins import (
    DictionaryData,
    SplitWordsFunction,
    TemporaryDictionaryMixin,
)
from montreal_forced_aligner.exceptions import DictionaryError, DictionaryFileError
from montreal_forced_aligner.models import DictionaryModel
from montreal_forced_aligner.utils import thirdparty_binary

__all__ = [
    "PronunciationDictionaryMixin",
]


class PronunciationDictionaryMixin(TemporaryDictionaryMixin):
    """
    Abstract mixin class containing information about a pronunciation dictionary

    Parameters
    ----------
    dictionary_path : str
        Path to pronunciation dictionary
    root_dictionary : :class:`~montreal_forced_aligner.dictionary.mixins.TemporaryDictionaryMixin`, optional
        Optional root dictionary to take phone information from

    See Also
    --------
    :class:`~montreal_forced_aligner.dictionary.mixins.DictionaryMixin`
        For dictionary parsing parameters
    :class:`~montreal_forced_aligner.abc.TemporaryDirectoryMixin`
        For temporary directory parameters

    Attributes
    ----------
    dictionary_model: DictionaryModel
        Dictionary model to load
    words: WordsType
        Words mapped to their pronunciations
    graphemes: set[str]
        Set of graphemes in the dictionary
    words_mapping: MappingType
        Mapping of words to integer IDs
    lexicon_word_set: set[str]
        Word set to limit output of lexicon files
    """

    def __init__(self, dictionary_path, root_dictionary=None, **kwargs):
        super().__init__(**kwargs)
        self.dictionary_model = DictionaryModel(dictionary_path)
        self.root_dictionary = root_dictionary
        os.makedirs(self.dictionary_output_directory, exist_ok=True)
        self.words = {}
        self.graphemes = set()
        self.words[self.silence_word] = [
            {"pronunciation": (self.nonoptional_silence_phone,), "probability": 1}
        ]
        self.words[self.oov_word] = [{"pronunciation": (self.oov_phone,), "probability": 1}]
        self.lookup_cache = {}
        self.int_cache = {}
        self.check_cache = {}
        with open(self.dictionary_model.path, "r", encoding="utf8") as inf:
            for i, line in enumerate(inf):
                line = line.strip()
                if not line:
                    continue
                line = line.split()
                word = self.sanitize(line.pop(0).lower())
                if not line:
                    raise DictionaryError(
                        f"Line {i} of {self.dictionary_model.path} does not have a pronunciation."
                    )
                if word in [self.silence_word, self.oov_word]:
                    continue
                self.graphemes.update(word)
                prob = 1
                if self.dictionary_model.pronunciation_probabilities:
                    prob = float(line.pop(0))
                    if prob > 1 or prob < 0:
                        raise ValueError
                if self.dictionary_model.silence_probabilities:
                    right_sil_prob = float(line.pop(0))
                    left_sil_prob = float(line.pop(0))
                    left_nonsil_prob = float(line.pop(0))
                else:
                    right_sil_prob = None
                    left_sil_prob = None
                    left_nonsil_prob = None
                if self.multilingual_ipa:
                    pron = self.parse_ipa(line)
                else:
                    pron = tuple(line)
                pronunciation = {
                    "pronunciation": pron,
                    "probability": prob,
                    "disambiguation": None,
                    "right_sil_prob": right_sil_prob,
                    "left_sil_prob": left_sil_prob,
                    "left_nonsil_prob": left_nonsil_prob,
                }
                if self.multilingual_ipa:
                    pronunciation["original_pronunciation"] = tuple(line)
                if not any(x in self.silence_phones for x in pron):
                    self.non_silence_phones.update(pron)
                if word in self.words and pron in {x["pronunciation"] for x in self.words[word]}:
                    continue
                if word not in self.words:
                    self.words[word] = []
                self.words[word].append(pronunciation)
                # test whether a word is a clitic
                is_clitic = False
                for cm in self.clitic_markers:
                    if word.startswith(cm) or word.endswith(cm):
                        is_clitic = True
                if is_clitic:
                    self.clitic_set.add(word)
        self.words_mapping = {}
        self._dictionary_data: Optional[DictionaryData] = None
        self.lexicon_word_set = None
        if not self.graphemes:
            raise DictionaryFileError(
                f"No words were found in the dictionary path {self.dictionary_model.path}"
            )

    @property
    def name(self) -> str:
        """Name of the dictionary"""
        return self.dictionary_model.name

    def __hash__(self) -> Any:
        """Return the hash of a given dictionary"""
        return hash(self.dictionary_model.path)

    @property
    def dictionary_output_directory(self) -> str:
        """Temporary directory to store all dictionary information"""
        return os.path.join(self.temporary_directory, self.name)

    @property
    def silences(self) -> set[str]:
        """
        Set of symbols that correspond to silence
        """
        return self.silence_phones

    def data(self, word_set: Optional[Collection[str]] = None) -> DictionaryData:
        """
        Generates a dictionary data for use in parsing utilities

        Parameters
        ----------
        word_set: Collection[str], optional
            Word set to limit data to

        Returns
        -------
        DictionaryData
            Data necessary for parsing text
        """

        def word_check(word):
            """Check whether a word should be included in the output"""
            if word in word_set:
                return True
            if word in self.clitic_set:
                return True
            if word in self.specials_set:
                return True
            return False

        if word_set:
            words_mapping = {k: v for k, v in self.words_mapping.items() if word_check(k)}
            reversed_word_mapping = {
                k: v for k, v in self.reversed_word_mapping.items() if word_check(v)
            }
            words = {k: v for k, v in self.words.items() if word_check(k)}
        else:
            words_mapping = self.words_mapping
            reversed_word_mapping = self.reversed_word_mapping
            words = self.words
        return DictionaryData(
            self.dictionary_options,
            self.construct_sanitize_function(),
            self.construct_split_words_function(),
            words_mapping,
            reversed_word_mapping,
            words,
            self.lookup_cache,
        )

    def construct_split_words_function(self) -> SplitWordsFunction:
        """
        Construct a :class:`~montreal_forced_aligner.dictionary.mixins.SplitWordsFunction` to use in multiprocessing jobs

        Returns
        -------
        :class:`~montreal_forced_aligner.dictionary.mixins.SplitWordsFunction`
            Function for sanitizing text
        """
        f = SplitWordsFunction(
            self.punctuation,
            self.clitic_markers,
            self.compound_markers,
            self.brackets,
            self.clitic_set,
            set(self.words.keys()),
        )

        return f

    def set_lexicon_word_set(self, word_set: Collection[str]) -> None:
        """
        Limit lexicon output to a subset of overall words

        Parameters
        ----------
        word_set: Collection[str]
            Word set to limit generated files to
        """
        split_function = self.construct_split_words_function()
        self.lexicon_word_set = {self.silence_word, self.oov_word}
        self.lexicon_word_set.update(self.clitic_set)
        for word in word_set:
            self.lookup_cache[word] = split_function(word)
            self.lexicon_word_set.update(self.lookup_cache[word])

        self._dictionary_data = self.data(self.lexicon_word_set)
        self.generate_mappings()

    def split_clitics(self, item: str) -> list[str]:
        """
        Split a word into subwords based on clitic and compound markers

        Parameters
        ----------
        item: str
            Word to split up

        Returns
        -------
        list[str]
            List of subwords
        """

        return self.construct_split_words_function()(item)

    def __bool__(self):
        """Check that the dictionary contains words"""
        return bool(self.words)

    def __len__(self) -> int:
        """Return the number of pronunciations across all words"""
        return sum(len(x) for x in self.words.values())

    def exclude_for_alignment(self, word: str) -> bool:
        """
        Check for whether to exclude a word from alignment lexicons (if there is a word set in the dictionary,
        checks whether the given string is in the word set)

        Parameters
        ----------
        word: str
            Word to check

        Returns
        -------
        bool
            True if there is no word set on the dictionary, or if the word is in the given word set
        """
        if self.lexicon_word_set is None:
            return False
        if word not in self.lexicon_word_set and word not in self.clitic_set:
            return True
        return False

    def generate_mappings(self) -> None:
        """
        Generate word mappings from text to integer IDs
        """
        self.words_mapping = {}
        i = 0
        self.words_mapping["<eps>"] = i
        for w in sorted(self.words.keys()):
            if self.exclude_for_alignment(w):
                continue
            i += 1
            self.words_mapping[w] = i

        self.words_mapping["#0"] = i + 1
        self.words_mapping["<s>"] = i + 2
        self.words_mapping["</s>"] = i + 3
        self.oovs_found = Counter()
        self.add_disambiguation()
        self._dictionary_data = self.data()

    def add_disambiguation(self) -> None:
        """
        Calculate disambiguation symbols for each pronunciation
        """
        subsequences = set()
        pronunciation_counts = defaultdict(int)

        for w, prons in self.words.items():
            if self.exclude_for_alignment(w):
                continue
            for p in prons:
                pronunciation_counts[p["pronunciation"]] += 1
                pron = p["pronunciation"][:-1]
                while pron:
                    subsequences.add(tuple(p))
                    pron = pron[:-1]
        last_used = defaultdict(int)
        for w, prons in sorted(self.words.items()):
            if self.exclude_for_alignment(w):
                continue
            for p in prons:
                if (
                    pronunciation_counts[p["pronunciation"]] == 1
                    and not p["pronunciation"] in subsequences
                ):
                    disambig = None
                else:
                    pron = p["pronunciation"]
                    last_used[pron] += 1
                    disambig = last_used[pron]
                p["disambiguation"] = disambig
        if last_used:
            self.max_disambiguation_symbol = max(
                self.max_disambiguation_symbol, max(last_used.values())
            )

    def create_utterance_fst(self, text: list[str], frequent_words: list[tuple[str, int]]) -> str:
        """
        Create an FST for an utterance with frequent words as a unigram language model

        Parameters
        ----------
        text: list[str]
            Text of the utterance
        frequent_words: list[tuple[str, int]]
            Frequent words to incorporate into the FST
        Returns
        -------
        str
            FST created from the utterance text and frequent words
        """
        num_words = len(text)
        word_probs = Counter(text)
        word_probs = {k: v / num_words for k, v in word_probs.items()}
        word_probs.update(frequent_words)
        fst_text = ""
        for k, v in word_probs.items():
            cost = -1 * math.log(v)
            w = self.to_int(k)[0]
            fst_text += f"0 0 {w} {w} {cost}\n"
        fst_text += f"0 {-1 * math.log(1 / num_words)}\n"
        return fst_text

    def to_int(self, item: str) -> list[int]:
        """
        Convert a given word into integer IDs

        Parameters
        ----------
        item: str
            Word to look up

        Returns
        -------
        list[int]
            List of integer IDs corresponding to each subword
        """
        if item not in self.int_cache:
            self.int_cache[item] = self._dictionary_data.to_int(item)
        return self.int_cache[item]

    def _lookup(self, item: str) -> list[str]:
        """
        Look up a word and return the list of sub words if necessary taking into account clitic and compound markers

        Parameters
        ----------
        item: str
            Word to look up

        Returns
        -------
        list[str]
            List of subwords that are in the dictionary
        """
        if item not in self.lookup_cache:
            if self._dictionary_data is not None:
                self.lookup_cache[item] = self._dictionary_data.lookup(item)
            else:
                self.lookup_cache[item] = self.construct_split_words_function()(item)
        return self.lookup_cache[item]

    def check_word(self, item: str) -> bool:
        """
        Check whether a word is in the dictionary, takes into account sanitization and
        clitic and compound markers

        Parameters
        ----------
        item: str
            Word to check

        Returns
        -------
        bool
            True if the look up would not result in an OOV item
        """
        if item not in self.check_cache:
            self.check_cache[item] = self._dictionary_data.check_word(item)
        return self.check_cache[item]

    @property
    def reversed_word_mapping(self) -> ReversedMappingType:
        """
        A mapping of integer ids to words
        """
        mapping = {}
        for k, v in self.words_mapping.items():
            mapping[v] = k
        return mapping

    @property
    def reversed_phone_mapping(self) -> ReversedMappingType:
        """
        A mapping of integer ids to phones
        """
        mapping = {}
        for k, v in self.phone_mapping.items():
            mapping[v] = k
        return mapping

    @property
    def oov_int(self) -> int:
        """
        The integer id for out of vocabulary items
        """
        return self.words_mapping[self.oov_word]

    @property
    def phones_dir(self) -> str:
        """
        Directory to store information Kaldi needs about phones
        """
        if self.root_dictionary is not None:
            return self.root_dictionary.phones_dir

        return os.path.join(self.dictionary_output_directory, "phones")

    @property
    def words_symbol_path(self) -> str:
        """
        Path of word to int mapping file for the dictionary
        """
        return os.path.join(self.dictionary_output_directory, "words.txt")

    @property
    def lexicon_fst_path(self) -> str:
        """
        Path of disambiguated lexicon fst (L.fst)
        """
        return os.path.join(self.dictionary_output_directory, "L.fst")

    @property
    def lexicon_disambig_fst_path(self) -> str:
        """
        Path of disambiguated lexicon fst (L.fst)
        """
        return os.path.join(self.dictionary_output_directory, "L_disambig.fst")

    def write(
        self,
        write_disambiguation: bool = False,
        debug=False,
    ) -> None:
        """
        Write the files necessary for Kaldi

        Parameters
        ----------
        write_disambiguation: bool, optional
            Flag for including disambiguation information
        debug: bool, optional
            Flag for whether to keep temporary files, defaults to False
        """
        if self.root_dictionary is None:
            self.generate_mappings()
            os.makedirs(self.phones_dir, exist_ok=True)
            self._write_word_boundaries()
            self._write_phone_map_file()
            self._write_phone_sets()
            self._write_phone_symbol_table()
            self._write_disambig()
            self._write_topo()
            self._write_extra_questions()

        self._write_graphemes()
        self._write_word_file()
        self._write_align_lexicon()
        if write_disambiguation:
            self._write_fst_text_disambiguated()
        else:
            self._write_basic_fst_text()
        self._write_fst_binary(write_disambiguation=write_disambiguation)
        if not debug:
            self.cleanup()

    def cleanup(self) -> None:
        """
        Clean up temporary files in the output directory
        """
        if os.path.exists(os.path.join(self.dictionary_output_directory, "temp.fst")):
            os.remove(os.path.join(self.dictionary_output_directory, "temp.fst"))
        if os.path.exists(os.path.join(self.dictionary_output_directory, "lexicon.text.fst")):
            os.remove(os.path.join(self.dictionary_output_directory, "lexicon.text.fst"))

    def _write_graphemes(self) -> None:
        """
        Write graphemes to temporary directory
        """
        outfile = os.path.join(self.dictionary_output_directory, "graphemes.txt")
        with open(outfile, "w", encoding="utf8") as f:
            for char in sorted(self.graphemes):
                f.write(char + "\n")

    def export_lexicon(
        self,
        path: str,
        write_disambiguation: Optional[bool] = False,
        probability: Optional[bool] = False,
    ) -> None:
        """
        Export pronunciation dictionary to a text file

        Parameters
        ----------
        path: str
            Path to save dictionary
        write_disambiguation: bool, optional
            Flag for whether to include disambiguation information
        probability: bool, optional
            Flag for whether to include probabilities
        """
        with open(path, "w", encoding="utf8") as f:
            for w in sorted(self.words.keys()):
                for p in sorted(
                    self.words[w],
                    key=lambda x: (x["pronunciation"], x["probability"], x["disambiguation"]),
                ):
                    phones = " ".join(p["pronunciation"])
                    if write_disambiguation and p["disambiguation"] is not None:
                        phones += f" #{p['disambiguation']}"
                    if probability:
                        f.write(f"{w}\t{p['probability']}\t{phones}\n")
                    else:
                        f.write(f"{w}\t{phones}\n")

    def _write_word_file(self) -> None:
        """
        Write the word mapping to the temporary directory
        """
        words_path = os.path.join(self.dictionary_output_directory, "words.txt")
        if sys.platform == "win32":
            newline = ""
        else:
            newline = None
        with open(words_path, "w", encoding="utf8", newline=newline) as f:
            for w, i in sorted(self.words_mapping.items(), key=lambda x: x[1]):
                f.write(f"{w} {i}\n")

    def _write_align_lexicon(self) -> None:
        """
        Write the alignment lexicon text file to the temporary directory
        """
        if self.root_dictionary is None:
            phone_mapping = self.phone_mapping
        else:
            phone_mapping = self.root_dictionary.phone_mapping
        path = os.path.join(self.phones_dir, "align_lexicon.int")
        if os.path.exists(path):
            return

        with open(path, "w", encoding="utf8") as f:
            for w, i in self.words_mapping.items():
                if self.exclude_for_alignment(w):
                    continue
                if w not in self.words:  # special characters
                    continue
                for pron in sorted(
                    self.words[w],
                    key=lambda x: (x["pronunciation"], x["probability"], x["disambiguation"]),
                ):

                    phones = list(pron["pronunciation"])
                    if self.position_dependent_phones:
                        if len(phones) == 1:
                            phones[0] += "_S"
                        else:
                            for j in range(len(phones)):
                                if j == 0:
                                    phones[j] += "_B"
                                elif j == len(phones) - 1:
                                    phones[j] += "_E"
                                else:
                                    phones[j] += "_I"
                    p = " ".join(str(phone_mapping[x]) for x in phones)
                    f.write(f"{i} {i} {p}\n".format(i=i, p=p))

    def _write_fst_binary(
        self,
        write_disambiguation: Optional[bool] = False,
    ) -> None:
        """
        Write the binary fst file to the temporary directory

        See Also
        --------
        :kaldi_src:`fstaddselfloops`
            Relevant Kaldi binary
        :openfst_src:`fstcompile`
            Relevant OpenFst binary
        :openfst_src:`fstarcsort`
            Relevant OpenFst binary

        Parameters
        ----------
        write_disambiguation: bool, optional
            Flag for including disambiguation symbols
        """
        if write_disambiguation:
            lexicon_fst_path = os.path.join(
                self.dictionary_output_directory, "lexicon_disambig.text.fst"
            )
            output_fst = os.path.join(self.dictionary_output_directory, "L_disambig.fst")
        else:
            lexicon_fst_path = os.path.join(self.dictionary_output_directory, "lexicon.text.fst")
            output_fst = os.path.join(self.dictionary_output_directory, "L.fst")
        if self.root_dictionary is not None:
            phone_mapping = self.root_dictionary.phone_mapping
            phones_file_path = self.root_dictionary.phone_symbol_table_path
        else:
            phone_mapping = self.phone_mapping
            phones_file_path = self.phone_symbol_table_path
        words_file_path = os.path.join(self.dictionary_output_directory, "words.txt")

        log_path = os.path.join(self.dictionary_output_directory, "fst.log")
        temp_fst_path = os.path.join(self.dictionary_output_directory, "temp.fst")
        with open(log_path, "w") as log_file:
            compile_proc = subprocess.Popen(
                [
                    thirdparty_binary("fstcompile"),
                    f"--isymbols={phones_file_path}",
                    f"--osymbols={words_file_path}",
                    "--keep_isymbols=false",
                    "--keep_osymbols=false",
                    lexicon_fst_path,
                    temp_fst_path,
                ],
                stderr=log_file,
            )
            compile_proc.communicate()
            if write_disambiguation:
                temp2_fst_path = os.path.join(self.dictionary_output_directory, "temp2.fst")
                word_disambig_path = os.path.join(
                    self.dictionary_output_directory, "word_disambig0.txt"
                )
                phone_disambig_path = os.path.join(
                    self.dictionary_output_directory, "phone_disambig0.txt"
                )
                with open(phone_disambig_path, "w") as f:
                    f.write(str(phone_mapping["#0"]))
                with open(word_disambig_path, "w") as f:
                    f.write(str(self.words_mapping["#0"]))
                selfloop_proc = subprocess.Popen(
                    [
                        thirdparty_binary("fstaddselfloops"),
                        phone_disambig_path,
                        word_disambig_path,
                        temp_fst_path,
                        temp2_fst_path,
                    ],
                    stderr=log_file,
                )
                selfloop_proc.communicate()
                arc_sort_proc = subprocess.Popen(
                    [
                        thirdparty_binary("fstarcsort"),
                        "--sort_type=olabel",
                        temp2_fst_path,
                        output_fst,
                    ],
                    stderr=log_file,
                )
            else:
                arc_sort_proc = subprocess.Popen(
                    [
                        thirdparty_binary("fstarcsort"),
                        "--sort_type=olabel",
                        temp_fst_path,
                        output_fst,
                    ],
                    stderr=log_file,
                )
            arc_sort_proc.communicate()

    def _write_basic_fst_text(self) -> None:
        """
        Write the L.fst text file to the temporary directory
        """
        nonoptional_silence = None
        optional_silence_phone = None
        lexicon_fst_path = os.path.join(self.dictionary_output_directory, "lexicon.text.fst")
        start_state = 0
        silence_state = 0
        silence_cost = 0
        no_silence_cost = 0
        loop_state = 0
        next_state = 1
        if self.silence_probability:
            optional_silence_phone = self.optional_silence_phone
            nonoptional_silence = self.nonoptional_silence_phone

            silence_cost = -1 * math.log(self.silence_probability)
            no_silence_cost = -1 * math.log(1.0 - self.silence_probability)
            loop_state = 1
            silence_state = 2

        with open(lexicon_fst_path, "w", encoding="utf8") as outf:
            if self.silence_probability:
                outf.write(
                    "\t".join(
                        map(str, [start_state, loop_state, "<eps>", "<eps>", no_silence_cost])
                    )
                    + "\n"
                )  # no silence

                outf.write(
                    "\t".join(
                        map(
                            str,
                            [start_state, loop_state, nonoptional_silence, "<eps>", silence_cost],
                        )
                    )
                    + "\n"
                )  # silence
                outf.write(
                    "\t".join(
                        map(str, [silence_state, loop_state, optional_silence_phone, "<eps>"])
                    )
                    + "\n"
                )  # no cost
                next_state = 3

            for w in sorted(self.words.keys()):
                if self.exclude_for_alignment(w):
                    continue
                for pron in sorted(
                    self.words[w],
                    key=lambda x: (x["pronunciation"], x["probability"], x["disambiguation"]),
                ):
                    phones = list(pron["pronunciation"])
                    prob = pron["probability"]
                    if self.position_dependent_phones:
                        if len(phones) == 1:
                            phones[0] += "_S"
                        else:
                            for i in range(len(phones)):
                                if i == 0:
                                    phones[i] += "_B"
                                elif i == len(phones) - 1:
                                    phones[i] += "_E"
                                else:
                                    phones[i] += "_I"
                    if not prob:
                        prob = 0.001  # Dithering to ensure low probability entries
                    pron_cost = abs(math.log(prob))

                    current_state = loop_state
                    word_or_eps = w
                    local_no_silence_cost = no_silence_cost + pron_cost
                    local_silcost = no_silence_cost + pron_cost
                    for i, p in enumerate(phones):
                        if i < len(phones) - 1:
                            outf.write(
                                f"{current_state}\t{next_state}\t{p}\t{word_or_eps}\t{pron_cost}\n"
                            )
                            word_or_eps = "<eps>"
                            pron_cost = 0
                            current_state = next_state
                            next_state += 1
                        else:  # transition on last phone to loop state
                            if self.silence_probability:
                                outf.write(
                                    f"{current_state}\t{loop_state}\t{p}\t{word_or_eps}\t{local_no_silence_cost}\n"
                                )
                                outf.write(
                                    f"{current_state}\t{silence_state}\t{p}\t{word_or_eps}\t{local_silcost}\n"
                                )
                            else:
                                outf.write(
                                    f"{current_state}\t{loop_state}\t{p}\t{word_or_eps}\t{pron_cost}\n"
                                )
                                word_or_eps = "<eps>"

            outf.write(f"{loop_state}\t0\n")

    def _write_fst_text_disambiguated(
        self, multispeaker_dictionary: Optional[MultispeakerDictionaryMixin] = None
    ) -> None:
        """
        Write the text L_disambig.fst file to the temporary directory

        Parameters
        ----------
        multispeaker_dictionary: MultispeakerDictionaryMixin, optional
            Main dictionary with phone mappings
        """
        lexicon_fst_path = os.path.join(
            self.dictionary_output_directory, "lexicon_disambig.text.fst"
        )
        if multispeaker_dictionary is not None:
            sil_disambiguation = f"#{multispeaker_dictionary.max_disambiguation_symbol + 1}"
        else:
            sil_disambiguation = f"#{self.max_disambiguation_symbol + 1}"
        assert self.silence_probability
        start_state = 0
        loop_state = 1
        silence_state = 2
        next_state = 3

        silence_phone = self.nonoptional_silence_phone

        silence_cost = -1 * math.log(self.silence_probability)
        no_silence_cost = -1 * math.log(1 - self.silence_probability)

        with open(lexicon_fst_path, "w", encoding="utf8") as outf:
            outf.write(
                f"{start_state}\t{loop_state}\t<eps>\t<eps>\t{no_silence_cost}\n"
            )  # no silence
            outf.write(
                f"{start_state}\t{silence_state}\t<eps>\t<eps>\t{silence_cost}\n"
            )  # silence
            silence_disambiguation_state = next_state
            next_state += 1

            outf.write(
                f"{silence_state}\t{silence_disambiguation_state}\t{silence_phone}\t<eps>\t0.0\n"
            )  # silence disambig
            outf.write(
                f"{silence_disambiguation_state}\t{loop_state}\t{sil_disambiguation}\t<eps>\t0.0\n"
            )  # silence disambig

            for w in sorted(self.words.keys()):
                if self.exclude_for_alignment(w):
                    continue
                for pron in sorted(
                    self.words[w],
                    key=lambda x: (x["pronunciation"], x["probability"], x["disambiguation"]),
                ):
                    phones = list(pron["pronunciation"])
                    prob = pron["probability"]
                    disambig_symbol = pron["disambiguation"]
                    if self.position_dependent_phones:
                        if len(phones) == 1:
                            phones[0] += "_S"
                        else:
                            for i in range(len(phones)):
                                if i == 0:
                                    phones[i] += "_B"
                                elif i == len(phones) - 1:
                                    phones[i] += "_E"
                                else:
                                    phones[i] += "_I"
                    if not prob:
                        prob = 0.001  # Dithering to ensure low probability entries
                    pron_cost = abs(math.log(prob))
                    if disambig_symbol:
                        phones += [f"#{disambig_symbol}"]

                    current_state = loop_state
                    for i in range(0, len(phones) - 1):
                        p = phones[i]
                        outf.write(
                            f"{current_state}\t{next_state}\t{p}\t{w if i == 0 else '<eps>'}\t{pron_cost if i == 0 else 0.0}\n"
                        )
                        current_state = next_state
                        next_state += 1

                    i = len(phones) - 1

                    local_no_silence_cost = no_silence_cost + pron_cost
                    local_silcost = silence_cost + pron_cost
                    if i <= 0:
                        local_silcost = silence_cost
                        local_no_silence_cost = no_silence_cost
                    outf.write(
                        f"{current_state}\t{loop_state}\t{phones[i] if i >= 0 else '<eps>'}\t{w if i <= 0 else '<eps>'}\t{local_no_silence_cost}\n"
                    )
                    outf.write(
                        f"{current_state}\t{silence_state}\t{phones[i] if i >= 0 else '<eps>'}\t{w if i <= 0 else '<eps>'}\t{local_silcost}\n"
                    )

            outf.write(f"{loop_state}\t0.0\n")


class PronunciationDictionary(PronunciationDictionaryMixin):
    """
    Class for processing pronunciation dictionaries

    See Also
    --------
    :class:`~montreal_forced_aligner.dictionary.pronunciation.PronunciationDictionaryMixin`
        For acoustic model training parsing parameters
    """

    @property
    def data_source_identifier(self) -> str:
        """Dictionary name"""
        return f"{self.name}"

    @property
    def identifier(self) -> str:
        """Dictionary name"""
        return f"{self.data_source_identifier}"

    @property
    def output_directory(self) -> str:
        """Temporary directory for the dictionary"""
        return os.path.join(self.temporary_directory, self.identifier)



"""
Textgrid utilities
==================

"""
from __future__ import annotations

import os
from typing import TYPE_CHECKING, Optional

from praatio import textgrid as tgio

from .data import CtmInterval

if TYPE_CHECKING:
    from .abc import ReversedMappingType
    from .alignment.base import CorpusAligner
    from .corpus.classes import File, Speaker
    from .dictionary import DictionaryData

__all__ = [
    "process_ctm_line",
    "parse_from_word",
    "parse_from_phone",
    "parse_from_word_no_cleanup",
    "generate_tiers",
    "export_textgrid",
    "ctm_to_textgrid",
    "output_textgrid_writing_errors",
]


def process_ctm_line(line: str) -> CtmInterval:
    """
    Helper function for parsing a line of CTM file to construct a CTMInterval

    Parameters
    ----------
    line: str
        Input string

    Returns
    -------
    :class:`~montreal_forced_aligner.data.CtmInterval`
        Extracted data from the line
    """
    line = line.split(" ")
    utt = line[0]
    if len(line) == 5:
        begin = round(float(line[2]), 4)
        duration = float(line[3])
        end = round(begin + duration, 4)
        label = line[4]
    else:
        begin = round(float(line[1]), 4)
        duration = float(line[2])
        end = round(begin + duration, 4)
        label = line[3]
    return CtmInterval(begin, end, label, utt)


def parse_from_word(
    ctm_labels: list[CtmInterval], text: list[str], dictionary_data: DictionaryData
) -> list[CtmInterval]:
    """
    Parse CTM intervals into the corresponding text for an utterance

    Parameters
    ----------
    ctm_labels: list[:class:`~montreal_forced_aligner.data.CtmInterval`]
        CTM intervals
    text: list[str]
        The original text that was to be aligned
    dictionary_data: :class:`~montreal_forced_aligner.dictionary.DictionaryData`
        Dictionary data necessary for splitting subwords

    Returns
    -------
    list[:class:`~montreal_forced_aligner.data.CtmInterval`]
        Correct intervals with subwords merged back into their original text
    """
    cur_ind = 0
    actual_labels = []
    utterance = None
    for word in text:
        ints = dictionary_data.to_int(word)
        b = 1000000
        e = -1
        for i in ints:
            cur = ctm_labels[cur_ind]
            if utterance is None:
                utterance = cur.utterance
            if i == int(cur.label):
                if cur.begin < b:
                    b = cur.begin
                if cur.end > e:
                    e = cur.end
            cur_ind += 1
        lab = CtmInterval(b, e, word, utterance)
        actual_labels.append(lab)
    return actual_labels


def parse_from_word_no_cleanup(
    ctm_labels: list[CtmInterval], reversed_word_mapping: ReversedMappingType
) -> list[CtmInterval]:
    """
    Assume that subwords in the CTM files are desired, so just does a reverse look up to get the sub word
    text

    Parameters
    ----------
    ctm_labels: list[:class:`~montreal_forced_aligner.data.CtmInterval`]
        List of :class:`~montreal_forced_aligner.data.CtmInterval` to convert
    reversed_word_mapping: dict[int, str]
        Look up for Kaldi word IDs to convert them back to text

    Returns
    -------
    list[:class:`~montreal_forced_aligner.data.CtmInterval`]
        Parsed intervals with text rather than integer IDs
    """
    for ctm_interval in ctm_labels:
        label = reversed_word_mapping[int(ctm_interval.label)]
        ctm_interval.label = label
    return ctm_labels


def parse_from_phone(
    ctm_labels: list[CtmInterval],
    reversed_phone_mapping: ReversedMappingType,
    positions: list[str],
) -> list[CtmInterval]:
    """
    Parse CtmIntervals to original phone transcriptions

    Parameters
    ----------
    ctm_labels: list[:class:`~montreal_forced_aligner.data.CtmInterval`]
        List of :class:`~montreal_forced_aligner.data.CtmInterval` to convert
    reversed_phone_mapping: dict[int, str]
        Mapping to convert phone IDs to phone labels
    positions: list[str]
        List of word positions to account for

    Returns
    -------
    list[:class:`~montreal_forced_aligner.data.CtmInterval`]
        Parsed intervals with phone labels rather than IDs
    """
    for ctm_interval in ctm_labels:
        label = reversed_phone_mapping[int(ctm_interval.label)]
        for p in positions:
            if label.endswith(p):
                label = label[: -1 * len(p)]
        ctm_interval.label = label
    return ctm_labels


def output_textgrid_writing_errors(output_directory: str, export_errors: dict[str, str]) -> None:
    """
    Output any errors that were encountered in writing TextGrids

    Parameters
    ----------
    output_directory: str
        Directory to save TextGrids files
    export_errors: dict[str, str]
        Dictionary of errors encountered
    """
    error_log = os.path.join(output_directory, "output_errors.txt")
    if os.path.exists(error_log):
        os.remove(error_log)
    for file_name, result in export_errors.items():
        if not os.path.exists(error_log):
            with open(error_log, "w", encoding="utf8") as f:
                f.write(
                    "The following exceptions were encountered during the output of the alignments to TextGrids:\n\n"
                )
        with open(error_log, "a", encoding="utf8") as f:
            f.write(f"{file_name}:\n")
            f.write(f"{result}\n\n")


def generate_tiers(
    file: File, cleanup_textgrids: Optional[bool] = True
) -> dict[Speaker, dict[str, list[CtmInterval]]]:
    """
    Generate TextGrid tiers for a given File

    Parameters
    ----------
    file: File
        File to generate TextGrid tiers
    cleanup_textgrids: bool, optional
        Flag for whether the TextGrids should be cleaned up

    Returns
    -------
    dict[Speaker, dict[str, list[:class:`~montreal_forced_aligner.data.CtmInterval`]]
        Tier information per speaker, with :class:`~montreal_forced_aligner.data.CtmInterval` split by "phones" and "words"
    """
    output = {}

    for u in file.utterances.values():
        if not u.word_labels:
            continue
        speaker = u.speaker
        dictionary_data = speaker.dictionary_data

        words = []
        phones = []
        if dictionary_data.multilingual_ipa and cleanup_textgrids:
            phone_ind = 0
            for interval in u.word_labels:
                end = interval.end
                word = interval.label
                subwords = dictionary_data.lookup(
                    word,
                )
                subwords = [
                    x if x in dictionary_data.words_mapping else dictionary_data.oov_word
                    for x in subwords
                ]
                subprons = [dictionary_data.words[x] for x in subwords]
                cur_phones = []
                while u.phone_labels[phone_ind].end <= end:
                    p = u.phone_labels[phone_ind]
                    if p.label in dictionary_data.silence_phones:
                        phone_ind += 1
                        continue
                    cur_phones.append(p)
                    phone_ind += 1
                    if phone_ind > len(u.phone_labels) - 1:
                        break
                phones.extend(dictionary_data.map_to_original_pronunciation(cur_phones, subprons))
                if not word:
                    continue

                words.append(interval)
        else:
            for interval in u.word_labels:
                words.append(interval)
            for interval in u.phone_labels:
                if interval.label in dictionary_data.silence_phones and cleanup_textgrids:
                    continue
                phones.append(interval)
        if speaker not in output:
            output[speaker] = {"words": words, "phones": phones}
        else:
            output[speaker]["words"].extend(words)
            output[speaker]["phones"].extend(phones)
    return output


def export_textgrid(
    file: File,
    output_path: str,
    speaker_data: dict[Speaker, dict[str, list[CtmInterval]]],
    frame_shift: int,
    first_file_write: Optional[bool] = True,
) -> None:
    """
    Export aligned file to TextGrid

    Parameters
    ----------
    file: File
        File object to export
    output_path: str
        Output path of the file
    speaker_data: dict[Speaker, dict[str, list[:class:`~montreal_forced_aligner.data.CtmInterval`]]
        Per speaker, per word/phone :class:`~montreal_forced_aligner.data.CtmInterval`
    frame_shift: int
        Frame shift of features, in ms
    first_file_write: bool, optional
        Flag for whether the file should be created from scratch or appended to if it
        has been modified by another export process
    """
    if frame_shift > 1:
        frame_shift = round(frame_shift / 1000, 4)
    if first_file_write:
        # Create initial textgrid
        tg = tgio.Textgrid()
        tg.minTimestamp = 0
        tg.maxTimestamp = file.duration

        if len(file.speaker_ordering) > 1:
            for speaker in file.speaker_ordering:
                word_tier_name = f"{speaker} - words"
                phone_tier_name = f"{speaker} - phones"

                word_tier = tgio.IntervalTier(word_tier_name, [], minT=0, maxT=file.duration)
                phone_tier = tgio.IntervalTier(phone_tier_name, [], minT=0, maxT=file.duration)
                tg.addTier(word_tier)
                tg.addTier(phone_tier)
        else:
            word_tier_name = "words"
            phone_tier_name = "phones"
            word_tier = tgio.IntervalTier(word_tier_name, [], minT=0, maxT=file.duration)
            phone_tier = tgio.IntervalTier(phone_tier_name, [], minT=0, maxT=file.duration)
            tg.addTier(word_tier)
            tg.addTier(phone_tier)
    else:
        # Use existing
        tg = tgio.openTextgrid(output_path, includeEmptyIntervals=False)

        word_tier_name = "words"
        phone_tier_name = "phones"
        word_tier = tgio.IntervalTier(word_tier_name, [], minT=0, maxT=file.duration)
        phone_tier = tgio.IntervalTier(phone_tier_name, [], minT=0, maxT=file.duration)
        tg.addTier(word_tier)
        tg.addTier(phone_tier)
    for speaker, data in speaker_data.items():
        words = data["words"]
        phones = data["phones"]
        tg_words = []
        tg_phones = []
        for w in words:
            if file.duration - w.end < frame_shift:  # Fix rounding issues
                w.end = file.duration
            tg_words.append(w.to_tg_interval())
        for p in phones:
            if file.duration - p.end < frame_shift:  # Fix rounding issues
                p.end = file.duration
            tg_phones.append(p.to_tg_interval())

        if len(file.speaker_ordering) > 1:
            word_tier_name = f"{speaker} - words"
            phone_tier_name = f"{speaker} - phones"
        else:
            word_tier_name = "words"
            phone_tier_name = "phones"
        word_tier = tgio.IntervalTier(word_tier_name, tg_words, minT=0, maxT=file.duration)
        phone_tier = tgio.IntervalTier(phone_tier_name, tg_phones, minT=0, maxT=file.duration)
        tg.replaceTier(word_tier_name, word_tier)
        tg.replaceTier(phone_tier_name, phone_tier)

    tg.save(output_path, includeBlankSpaces=True, format="long_textgrid", reportingMode="error")


def ctm_to_textgrid(file: File, aligner: CorpusAligner, first_file_write=True) -> None:
    """
    Export a File to TextGrid

    Parameters
    ----------
    file: File
        File to export
    aligner: CorpusAligner
        Aligner used to generate the alignments
    first_file_write: bool, optional
        Flag for whether this is the first time touching this file
    """
    data = generate_tiers(file, cleanup_textgrids=aligner.cleanup_textgrids)

    backup_output_directory = None
    if not aligner.overwrite:
        backup_output_directory = aligner.backup_output_directory
        os.makedirs(backup_output_directory, exist_ok=True)
    output_path = file.construct_output_path(aligner.textgrid_output, backup_output_directory)
    export_textgrid(file, output_path, data, aligner.frame_shift, first_file_write)



"""
Corpus loading worker
---------------------


"""
from __future__ import annotations

import multiprocessing as mp
import os
import sys
import traceback
from queue import Empty
from typing import TYPE_CHECKING, Collection, Optional, Union

from montreal_forced_aligner.exceptions import TextGridParseError, TextParseError
from montreal_forced_aligner.helper import output_mapping

if TYPE_CHECKING:

    from montreal_forced_aligner.abc import OneToManyMappingType, OneToOneMappingType
    from montreal_forced_aligner.corpus.helper import SoundFileInfoDict

    FileInfoDict = dict[
        str, Union[str, SoundFileInfoDict, OneToOneMappingType, OneToManyMappingType]
    ]
    from montreal_forced_aligner.abc import MappingType, ReversedMappingType, WordsType
    from montreal_forced_aligner.corpus.classes import File, Speaker, Utterance
    from montreal_forced_aligner.dictionary import DictionaryData, PronunciationDictionaryMixin
    from montreal_forced_aligner.utils import Stopped


__all__ = ["CorpusProcessWorker", "Job"]


class CorpusProcessWorker(mp.Process):
    """
    Multiprocessing corpus loading worker

    Attributes
    ----------
    job_q: :class:`~multiprocessing.Queue`
        Job queue for files to process
    return_dict: dict
        Dictionary to catch errors
    return_q: :class:`~multiprocessing.Queue`
        Return queue for processed Files
    stopped: :func:`~montreal_forced_aligner.utils.Stopped`
        Stop check for whether corpus loading should exit
    finished_adding: :class:`~montreal_forced_aligner.utils.Stopped`
        Signal that the main thread has stopped adding new files to be processed
    """

    def __init__(
        self,
        job_q: mp.Queue,
        return_dict: dict,
        return_q: mp.Queue,
        stopped: Stopped,
        finished_adding: Stopped,
    ):
        mp.Process.__init__(self)
        self.job_q = job_q
        self.return_dict = return_dict
        self.return_q = return_q
        self.stopped = stopped
        self.finished_adding = finished_adding

    def run(self) -> None:
        """
        Run the corpus loading job
        """
        from ..corpus.classes import parse_file

        while True:
            try:
                arguments = self.job_q.get(timeout=1)
            except Empty:
                if self.finished_adding.stop_check():
                    break
                continue
            self.job_q.task_done()
            if self.stopped.stop_check():
                continue
            try:
                file = parse_file(*arguments, stop_check=self.stopped.stop_check)
                self.return_q.put(file)
            except TextParseError as e:
                self.return_dict["decode_error_files"].append(e)
            except TextGridParseError as e:
                self.return_dict["textgrid_read_errors"][e.file_name] = e
            except Exception:
                self.stopped.stop()
                self.return_dict["error"] = arguments, Exception(
                    traceback.format_exception(*sys.exc_info())
                )
        return


class Job:
    """
    Class representing information about corpus jobs that will be run in parallel.
    Jobs have a set of speakers that they will process, along with all files and utterances associated with that speaker.
    As such, Jobs also have a set of dictionaries that the speakers use, and argument outputs are largely dependent on
    the pronunciation dictionaries in use.

    Parameters
    ----------
    name: int
        Job number is the job's identifier

    Attributes
    ----------
    speakers: list[:class:`~montreal_forced_aligner.corpus.classes.Speaker`]
        List of speakers associated with this job
    dictionaries: set[:class:`~montreal_forced_aligner.dictionary.PronunciationDictionary`]
        Set of dictionaries that the job's speakers use
    subset_utts: set[:class:`~montreal_forced_aligner.corpus.classes.Utterance`]
        When trainers are just using a subset of the corpus, the subset of utterances on each job will be set and used to
        filter the job's utterances
    subset_speakers: set[:class:`~montreal_forced_aligner.corpus.classes.Speaker`]
        When subset_utts is set, this property will be calculated as the subset of speakers that the utterances correspond to
    subset_dictionaries: set[:class:`~montreal_forced_aligner.dictionary.PronunciationDictionary`]
        Subset of dictionaries that the subset of speakers use

    """

    name: int
    speakers: list[Speaker]
    subset_utts: set[Utterance]
    subset_speakers: set[Speaker]
    dictionaries: set[PronunciationDictionaryMixin]
    subset_dictionaries: set[PronunciationDictionaryMixin]

    def __init__(self, name: int):
        self.name = name
        self.speakers = []
        self.dictionaries = set()

        self.subset_utts = set()
        self.subset_speakers = set()
        self.subset_dictionaries = set()

    def add_speaker(self, speaker: Speaker) -> None:
        """
        Add a speaker to a job

        Parameters
        ----------
        speaker: :class:`~montreal_forced_aligner.corpus.classes.Speaker`
            Speaker to add
        """
        self.speakers.append(speaker)
        self.dictionaries.add(speaker.dictionary)

    def set_subset(self, subset_utts: Optional[Collection[Utterance]]) -> None:
        """
        Set the current subset for the trainer

        Parameters
        ----------
        subset_utts: Collection[:class:`~montreal_forced_aligner.corpus.classes.Utterance`], optional
            Subset of utterances for this job to use
        """
        if subset_utts is None:
            self.subset_utts = set()
            self.subset_speakers = set()
            self.subset_dictionaries = set()
        else:
            self.subset_utts = set(u for u in subset_utts if u.speaker in self.speakers)
            self.subset_speakers = {u.speaker for u in subset_utts if u.speaker in self.speakers}
            self.subset_dictionaries = {s.dictionary for s in self.subset_speakers}

    def text_scp_data(self) -> dict[str, dict[str, list[str]]]:
        """
        Generate the job's data for Kaldi's text scp files

        Returns
        -------
        dict[str, dict[str, list[str]]]
            Text for each utterance, per dictionary name
        """
        data = {}
        utts = self.job_utts()
        for dict_name, utt_data in utts.items():
            data[dict_name] = {}
            for utt in utt_data.values():
                if not utt.text:
                    continue
                data[dict_name][utt.name] = " ".join(map(str, utt.text_for_scp()))
        return data

    def text_int_scp_data(self) -> dict[str, dict[str, str]]:
        """
        Generate the job's data for Kaldi's text int scp files

        Returns
        -------
        dict[str, dict[str, str]]
            Text converted to integer IDs for each utterance, per dictionary name
        """
        data = {}
        utts = self.job_utts()
        for dict_name, utt_data in utts.items():
            data[dict_name] = {}
            for utt in utt_data.values():
                if utt.speaker.dictionary is None:
                    continue
                if not utt.text:
                    continue
                data[dict_name][utt.name] = " ".join(map(str, utt.text_int_for_scp()))
        return data

    def wav_scp_data(self) -> dict[str, dict[str, str]]:
        """
        Generate the job's data for Kaldi's wav scp files

        Returns
        -------
        dict[str, dict[str, str]]
            Wav scp strings for each file, per dictionary name
        """
        data = {}
        done = {}
        utts = self.job_utts()
        for dict_name, utt_data in utts.items():
            data[dict_name] = {}
            done[dict_name] = set()
            for utt in utt_data.values():
                if not utt.is_segment:
                    data[dict_name][utt.name] = utt.file.for_wav_scp()
                elif utt.file.name not in done:
                    data[dict_name][utt.file.name] = utt.file.for_wav_scp()
                    done[dict_name].add(utt.file.name)
        return data

    def utt2spk_scp_data(self) -> dict[str, dict[str, str]]:
        """
        Generate the job's data for Kaldi's utt2spk scp files

        Returns
        -------
        dict[str, dict[str, str]]
            Utterance to speaker mapping, per dictionary name
        """
        data = {}
        utts = self.job_utts()
        for dict_name, utt_data in utts.items():
            data[dict_name] = {}
            for utt in utt_data.values():
                data[dict_name][utt.name] = utt.speaker_name
        return data

    def feat_scp_data(self) -> dict[str, dict[str, str]]:
        """
        Generate the job's data for Kaldi's feature scp files

        Returns
        -------
        dict[str, dict[str, str]]
            Utterance to feature archive ID mapping, per dictionary name
        """
        data = {}
        utts = self.job_utts()
        for dict_name, utt_data in utts.items():
            data[dict_name] = {}
            for utt in utt_data.values():
                if not utt.features:
                    continue
                data[dict_name][utt.name] = utt.features
        return data

    def spk2utt_scp_data(self) -> dict[str, dict[str, list[str]]]:
        """
        Generate the job's data for Kaldi's spk2utt scp files

        Returns
        -------
        dict[str, dict[str, list[str]]]
            Speaker to utterance mapping, per dictionary name
        """
        data = {}
        utts = self.job_utts()
        for dict_name, utt_data in utts.items():
            data[dict_name] = {}
            for utt in utt_data.values():
                if utt.speaker.name not in data[dict_name]:
                    data[dict_name][utt.speaker.name] = []
                data[dict_name][utt.speaker.name].append(str(utt))
        for k, v in data.items():
            for s, utts in v.items():
                data[k][s] = sorted(utts)
        return data

    def cmvn_scp_data(self) -> dict[str, dict[str, str]]:
        """
        Generate the job's data for Kaldi's CMVN scp files

        Returns
        -------
        dict[str, dict[str, str]]
            Speaker to CMVN mapping, per dictionary name
        """
        data = {}
        for s in self.speakers:
            if s.dictionary is None:
                key = None
            else:
                key = s.dictionary.name
            if key not in data:
                data[key] = {}
            if self.subset_speakers and s not in self.subset_speakers:
                continue
            if s.cmvn:
                data[key][s.name] = s.cmvn
        return data

    def segments_scp_data(self) -> dict[str, dict[str, str]]:
        """
        Generate the job's data for Kaldi's segments scp files

        Returns
        -------
        dict[str, dict[str, str]]
            Utterance to segment mapping, per dictionary name
        """
        data = {}
        utts = self.job_utts()
        for dict_name, utt_data in utts.items():
            data[dict_name] = {}
            for utt in utt_data.values():
                if not utt.is_segment:
                    continue
                data[dict_name][utt.name] = utt.segment_for_scp()
        return data

    def construct_path_dictionary(
        self, directory: str, identifier: str, extension: str
    ) -> dict[str, str]:
        """
        Helper function for constructing dictionary-dependent paths for the Job

        Parameters
        ----------
        directory: str
            Directory to use as the root
        identifier: str
            Identifier for the path name, like ali or acc
        extension: str
            Extension of the path, like .scp or .ark

        Returns
        -------
        dict[str, str]
            Path for each dictionary
        """
        output = {}
        for dict_name in self.current_dictionary_names:
            output[dict_name] = os.path.join(
                directory, f"{identifier}.{dict_name}.{self.name}.{extension}"
            )
        return output

    def construct_dictionary_dependent_paths(
        self, directory: str, identifier: str, extension: str
    ) -> dict[str, str]:
        """
        Helper function for constructing paths that depend only on the dictionaries of the job, and not the job name itself.
        These paths should be merged with all other jobs to get a full set of dictionary paths.

        Parameters
        ----------
        directory: str
            Directory to use as the root
        identifier: str
            Identifier for the path name, like ali or acc
        extension: str
            Extension of the path, like .scp or .ark

        Returns
        -------
        dict[str, str]
            Path for each dictionary
        """
        output = {}
        for dict_name in self.current_dictionary_names:
            output[dict_name] = os.path.join(directory, f"{identifier}.{dict_name}.{extension}")
        return output

    @property
    def dictionary_count(self):
        """Number of dictionaries currently used"""
        if self.subset_dictionaries:
            return len(self.subset_dictionaries)
        return len(self.dictionaries)

    @property
    def current_dictionaries(self) -> Collection[PronunciationDictionaryMixin]:
        """Current dictionaries depending on whether a subset is being used"""
        if self.subset_dictionaries:
            return self.subset_dictionaries
        return self.dictionaries

    @property
    def current_dictionary_names(self) -> list[Optional[str]]:
        """Current dictionary names depending on whether a subset is being used"""
        if self.subset_dictionaries:
            return sorted(x.name for x in self.subset_dictionaries)
        if self.dictionaries == {None}:
            return [None]
        return sorted(x.name for x in self.dictionaries)

    def word_boundary_int_files(self) -> dict[str, str]:
        """
        Generate mapping for dictionaries to word boundary int files

        Returns
        -------
        dict[str, str]
            Per dictionary word boundary int files
        """
        data = {}
        for dictionary in self.current_dictionaries:
            data[dictionary.name] = os.path.join(dictionary.phones_dir, "word_boundary.int")
        return data

    def reversed_phone_mappings(self) -> dict[str, ReversedMappingType]:
        """
        Generate mapping for dictionaries to reversed phone mapping

        Returns
        -------
        dict[str, ReversedMappingType]
            Per dictionary reversed phone mapping
        """
        data = {}
        for dictionary in self.current_dictionaries:
            data[dictionary.name] = dictionary.reversed_phone_mapping
        return data

    def reversed_word_mappings(self) -> dict[str, ReversedMappingType]:
        """
        Generate mapping for dictionaries to reversed word mapping

        Returns
        -------
        dict[str, ReversedMappingType]
            Per dictionary reversed word mapping
        """
        data = {}
        for dictionary in self.current_dictionaries:
            data[dictionary.name] = dictionary.reversed_word_mapping
        return data

    def words_mappings(self) -> dict[str, MappingType]:
        """
        Generate mapping for dictionaries to word mapping

        Returns
        -------
        dict[str, MappingType]
            Per dictionary word mapping
        """
        data = {}
        for dictionary in self.current_dictionaries:
            data[dictionary.name] = dictionary.words_mapping
        return data

    def words(self) -> dict[str, WordsType]:
        """
        Generate mapping for dictionaries to words

        Returns
        -------
        dict[str, WordsType]
            Per dictionary words
        """
        data = {}
        for dictionary in self.current_dictionaries:
            data[dictionary.name] = dictionary.words
        return data

    def punctuation(self):
        """
        Generate mapping for dictionaries to punctuation

        Returns
        -------
        dict[str, str]
            Per dictionary punctuation
        """
        data = {}
        for dictionary in self.current_dictionaries:
            data[dictionary.name] = dictionary.punctuation
        return data

    def clitic_set(self) -> dict[str, set[str]]:
        """
        Generate mapping for dictionaries to clitic sets

        Returns
        -------
        dict[str, str]
            Per dictionary clitic sets
        """
        data = {}
        for dictionary in self.current_dictionaries:
            data[dictionary.name] = dictionary.clitic_set
        return data

    def clitic_markers(self) -> dict[str, list[str]]:
        """
        Generate mapping for dictionaries to clitic markers

        Returns
        -------
        dict[str, str]
            Per dictionary clitic markers
        """
        data = {}
        for dictionary in self.current_dictionaries:
            data[dictionary.name] = dictionary.clitic_markers
        return data

    def compound_markers(self) -> dict[str, list[str]]:
        """
        Generate mapping for dictionaries to compound markers

        Returns
        -------
        dict[str, str]
            Per dictionary compound markers
        """
        data = {}
        for dictionary in self.current_dictionaries:
            data[dictionary.name] = dictionary.compound_markers
        return data

    def strip_diacritics(self) -> dict[str, list[str]]:
        """
        Generate mapping for dictionaries to diacritics to strip

        Returns
        -------
        dict[str, list[str]]
            Per dictionary strip diacritics
        """
        data = {}
        for dictionary in self.current_dictionaries:
            data[dictionary.name] = dictionary.strip_diacritics
        return data

    def oov_codes(self) -> dict[str, str]:
        """
        Generate mapping for dictionaries to oov symbols

        Returns
        -------
        dict[str, str]
            Per dictionary oov symbols
        """
        data = {}
        for dictionary in self.current_dictionaries:
            data[dictionary.name] = dictionary.oov_word
        return data

    def oov_ints(self) -> dict[str, int]:
        """
        Generate mapping for dictionaries to oov ints

        Returns
        -------
        dict[str, int]
            Per dictionary oov ints
        """
        data = {}
        for dictionary in self.current_dictionaries:
            data[dictionary.name] = dictionary.oov_int
        return data

    def positions(self) -> dict[str, list[str]]:
        """
        Generate mapping for dictionaries to positions

        Returns
        -------
        dict[str, list[str]]
            Per dictionary positions
        """
        data = {}
        for dictionary in self.current_dictionaries:
            data[dictionary.name] = dictionary.positions
        return data

    def silences(self) -> dict[str, set[str]]:
        """
        Generate mapping for dictionaries to silence symbols

        Returns
        -------
        dict[str, set[str]]
            Per dictionary silence symbols
        """
        data = {}
        for dictionary in self.current_dictionaries:
            data[dictionary.name] = dictionary.silences
        return data

    def multilingual_ipa(self) -> dict[str, bool]:
        """
        Generate mapping for dictionaries to multilingual IPA flags

        Returns
        -------
        dict[str, bool]
            Per dictionary multilingual IPA flags
        """
        data = {}
        for dictionary in self.current_dictionaries:
            data[dictionary.name] = dictionary.multilingual_ipa
        return data

    def job_utts(self) -> dict[str, dict[str, Utterance]]:
        """
        Generate utterances by dictionary name for the Job

        Returns
        -------
        dict[str, dict[str, :class:`~montreal_forced_aligner.corpus.classes.Utterance`]]
            Mapping of dictionary name to Utterance mappings
        """
        data = {}
        if self.subset_utts:
            utterances = self.subset_utts
        else:
            utterances = set()
            for s in self.speakers:
                utterances.update(s.utterances.values())
        for u in utterances:
            if u.ignored:
                continue
            if u.speaker.dictionary is None:
                dict_name = None
            else:
                dict_name = u.speaker.dictionary.name
            if dict_name not in data:
                data[dict_name] = {}
            data[dict_name][u.name] = u

        return data

    def job_files(self) -> dict[str, File]:
        """
        Generate files for the Job

        Returns
        -------
        dict[str, :class:`~montreal_forced_aligner.corpus.classes.File`]
            Mapping of file name to File objects
        """
        data = {}
        if self.subset_utts:
            utterances = self.subset_utts
        else:
            utterances = set()
            for s in self.speakers:
                utterances.update(s.utterances.values())
        for u in utterances:
            if u.ignored:
                continue
            data[u.file_name] = u.file
        return data

    def job_speakers(self) -> dict[str, Speaker]:
        """
        Generate files for the Job

        Returns
        -------
        dict[str, :class:`~montreal_forced_aligner.corpus.classes.Speaker`]
            Mapping of file name to File objects
        """
        data = {}
        if self.subset_speakers:
            speakers = self.subset_speakers
        else:
            speakers = self.speakers
        for s in speakers:
            data[s.name] = s
        return data

    def dictionary_data(self) -> dict[str, DictionaryData]:
        """
        Generate dictionary data for the job

        Returns
        -------
        dict[str, DictionaryData]
            Mapping of dictionary name to dictionary data
        """
        data = {}
        for dictionary in self.current_dictionaries:
            data[dictionary.name] = dictionary.data()
        return data

    def output_to_directory(self, split_directory: str) -> None:
        """
        Output job information to a directory

        Parameters
        ----------
        split_directory: str
            Directory to output to
        """
        wav = self.wav_scp_data()
        for dict_name, scp in wav.items():
            wav_scp_path = os.path.join(split_directory, f"wav.{dict_name}.{self.name}.scp")
            output_mapping(scp, wav_scp_path, skip_safe=True)

        spk2utt = self.spk2utt_scp_data()
        for dict_name, scp in spk2utt.items():
            spk2utt_scp_path = os.path.join(
                split_directory, f"spk2utt.{dict_name}.{self.name}.scp"
            )
            output_mapping(scp, spk2utt_scp_path)

        feats = self.feat_scp_data()
        for dict_name, scp in feats.items():
            feats_scp_path = os.path.join(split_directory, f"feats.{dict_name}.{self.name}.scp")
            output_mapping(scp, feats_scp_path)

        cmvn = self.cmvn_scp_data()
        for dict_name, scp in cmvn.items():
            cmvn_scp_path = os.path.join(split_directory, f"cmvn.{dict_name}.{self.name}.scp")
            output_mapping(scp, cmvn_scp_path)

        utt2spk = self.utt2spk_scp_data()
        for dict_name, scp in utt2spk.items():
            utt2spk_scp_path = os.path.join(
                split_directory, f"utt2spk.{dict_name}.{self.name}.scp"
            )
            output_mapping(scp, utt2spk_scp_path)

        segments = self.segments_scp_data()
        for dict_name, scp in segments.items():
            segments_scp_path = os.path.join(
                split_directory, f"segments.{dict_name}.{self.name}.scp"
            )
            output_mapping(scp, segments_scp_path)

        text_scp = self.text_scp_data()
        for dict_name, scp in text_scp.items():
            if not scp:
                continue
            text_scp_path = os.path.join(split_directory, f"text.{dict_name}.{self.name}.scp")
            output_mapping(scp, text_scp_path)

        text_int = self.text_int_scp_data()
        for dict_name, scp in text_int.items():
            if dict_name is None:
                continue
            if not scp:
                continue
            text_int_scp_path = os.path.join(
                split_directory, f"text.{dict_name}.{self.name}.int.scp"
            )
            output_mapping(scp, text_int_scp_path, skip_safe=True)



"""Class definitions for Monophone trainer"""
from __future__ import annotations

import os
import re
import subprocess
from typing import NamedTuple

from montreal_forced_aligner.acoustic_modeling.base import AcousticModelTrainingMixin
from montreal_forced_aligner.utils import run_mp, run_non_mp, thirdparty_binary


class MonoAlignEqualArguments(NamedTuple):
    """Arguments for :func:`~montreal_forced_aligner.acoustic_modeling.monophone.mono_align_equal_func`"""

    log_path: str
    dictionaries: list[str]
    feature_strings: dict[str, str]
    fst_scp_paths: dict[str, str]
    ali_ark_paths: dict[str, str]
    acc_paths: dict[str, str]
    model_path: str


def mono_align_equal_func(
    log_path: str,
    dictionaries: list[str],
    feature_strings: dict[str, str],
    fst_scp_paths: dict[str, str],
    ali_ark_paths: dict[str, str],
    acc_paths: dict[str, str],
    model_path: str,
):
    """
    Multiprocessing function for initializing monophone alignments

    See Also
    --------
    :meth:`.MonophoneTrainer.mono_align_equal`
        Main function that calls this function in parallel
    :meth:`.MonophoneTrainer.mono_align_equal_arguments`
        Job method for generating arguments for this function
    :kaldi_src:`align-equal-compiled`
        Relevant Kaldi binary
    :kaldi_src:`gmm-acc-stats-ali`
        Relevant Kaldi binary

    Parameters
    ----------
    log_path: str
        Path to save log output
    dictionaries: list[str]
        List of dictionary names
    feature_strings: dict[str, str]
        Dictionary of feature strings per dictionary name
    fst_scp_paths: dict[str, str]
        Dictionary of utterance FST scp files per dictionary name
    ali_ark_paths: dict[str, str]
        Dictionary of alignment archives per dictionary name
    acc_paths: dict[str, str]
        Dictionary of accumulated stats files per dictionary name
    model_path: str
        Path to the acoustic model file
    """
    with open(log_path, "w", encoding="utf8") as log_file:
        for dict_name in dictionaries:
            fst_path = fst_scp_paths[dict_name]
            ali_path = ali_ark_paths[dict_name]
            acc_path = acc_paths[dict_name]
            align_proc = subprocess.Popen(
                [
                    thirdparty_binary("align-equal-compiled"),
                    f"scp:{fst_path}",
                    feature_strings[dict_name],
                    f"ark:{ali_path}",
                ],
                stderr=log_file,
                env=os.environ,
            )
            align_proc.communicate()
            stats_proc = subprocess.Popen(
                [
                    thirdparty_binary("gmm-acc-stats-ali"),
                    "--binary=true",
                    model_path,
                    feature_strings[dict_name],
                    f"ark:{ali_path}",
                    acc_path,
                ],
                stdin=align_proc.stdout,
                stderr=log_file,
                env=os.environ,
            )
            stats_proc.communicate()


__all__ = ["MonophoneTrainer"]


class MonophoneTrainer(AcousticModelTrainingMixin):
    """
    Configuration class for monophone training

    Attributes
    ----------
    subset : int
        Number of utterances to use, defaults to 2000
    initial_gaussians : int
        Number of gaussians to begin training, defaults to 135
    max_gaussians : int
        Total number of gaussians, defaults to 1000
    power : float
        Exponent for number of gaussians according to occurrence counts, defaults to 0.25

    See Also
    --------
    :class:`~montreal_forced_aligner.acoustic_modeling.base.AcousticModelTrainingMixin`
        For acoustic model training parsing parameters
    """

    def __init__(
        self,
        subset: int = 2000,
        initial_gaussians: int = 135,
        max_gaussians: int = 1000,
        power: float = 0.25,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.subset = subset
        self.initial_gaussians = initial_gaussians
        self.max_gaussians = max_gaussians
        self.power = power

    def mono_align_equal_arguments(self) -> list[MonoAlignEqualArguments]:
        """
        Generate Job arguments for :func:`~montreal_forced_aligner.acoustic_modeling.monophone.mono_align_equal_func`

        Returns
        -------
        list[:class:`~montreal_forced_aligner.acoustic_modeling.monophone.MonoAlignEqualArguments`]
            Arguments for processing
        """
        feat_strings = self.worker.construct_feature_proc_strings()
        return [
            MonoAlignEqualArguments(
                os.path.join(self.working_log_directory, f"mono_align_equal.{j.name}.log"),
                j.current_dictionary_names,
                feat_strings[j.name],
                j.construct_path_dictionary(self.working_directory, "fsts", "scp"),
                j.construct_path_dictionary(self.working_directory, "ali", "ark"),
                j.construct_path_dictionary(self.working_directory, "0", "acc"),
                self.model_path,
            )
            for j in self.jobs
        ]

    def compute_calculated_properties(self) -> None:
        """Generate realignment iterations and initial gaussians based on configuration"""
        self.realignment_iterations = [0]
        for i in range(1, self.num_iterations):
            if i <= int(self.num_iterations / 4):
                self.realignment_iterations.append(i)
            elif i <= int(self.num_iterations * 2 / 4):
                if i - self.realignment_iterations[-1] > 1:
                    self.realignment_iterations.append(i)
            else:
                if i - self.realignment_iterations[-1] > 2:
                    self.realignment_iterations.append(i)

    @property
    def train_type(self) -> str:
        """Training identifier"""
        return "mono"

    @property
    def phone_type(self) -> str:
        """Phone type"""
        return "monophone"

    def mono_align_equal(self):
        """
        Multiprocessing function that creates equal alignments for base monophone training.

        See Also
        --------
        :func:`~montreal_forced_aligner.acoustic_modeling.monophone.mono_align_equal_func`
            Multiprocessing helper function for each job
        :meth:`.MonophoneTrainer.mono_align_equal_arguments`
            Job method for generating arguments for the helper function
        :kaldi_src:`gmm-sum-accs`
            Relevant Kaldi binary
        :kaldi_src:`gmm-est`
            Relevant Kaldi binary
        :kaldi_steps:`train_mono`
            Reference Kaldi script
        """

        arguments = self.mono_align_equal_arguments()

        if self.use_mp:
            run_mp(mono_align_equal_func, arguments, self.working_log_directory)
        else:
            run_non_mp(mono_align_equal_func, arguments, self.working_log_directory)

        log_path = os.path.join(self.working_log_directory, "update.0.log")
        with open(log_path, "w") as log_file:
            acc_files = []
            for x in arguments:
                acc_files.extend(sorted(x.acc_paths.values()))
            sum_proc = subprocess.Popen(
                [thirdparty_binary("gmm-sum-accs"), "-"] + acc_files,
                stderr=log_file,
                stdout=subprocess.PIPE,
                env=os.environ,
            )
            est_proc = subprocess.Popen(
                [
                    thirdparty_binary("gmm-est"),
                    "--min-gaussian-occupancy=3",
                    f"--mix-up={self.current_gaussians}",
                    f"--power={self.power}",
                    self.model_path,
                    "-",
                    self.next_model_path,
                ],
                stderr=log_file,
                stdin=sum_proc.stdout,
                env=os.environ,
            )
            est_proc.communicate()
        if not self.debug:
            for f in acc_files:
                os.remove(f)

    def _trainer_initialization(self) -> None:
        """Monophone training initialization"""
        self.iteration = 0
        tree_path = os.path.join(self.working_directory, "tree")

        feat_dim = self.worker.get_feat_dim()

        feature_string = self.worker.construct_base_feature_string()
        shared_phones_path = os.path.join(self.worker.phones_dir, "sets.int")
        init_log_path = os.path.join(self.working_log_directory, "init.log")
        temp_feats_path = os.path.join(self.working_directory, "temp_feats")
        with open(init_log_path, "w") as log_file:
            subprocess.call(
                [
                    thirdparty_binary("subset-feats"),
                    "--n=10",
                    feature_string,
                    f"ark:{temp_feats_path}",
                ],
                stderr=log_file,
            )
            subprocess.call(
                [
                    thirdparty_binary("gmm-init-mono"),
                    f"--shared-phones={shared_phones_path}",
                    f"--train-feats=ark:{temp_feats_path}",
                    os.path.join(self.worker.topo_path),
                    str(feat_dim),
                    self.model_path,
                    tree_path,
                ],
                stderr=log_file,
            )
            proc = subprocess.Popen(
                [thirdparty_binary("gmm-info"), "--print-args=false", self.model_path],
                stderr=log_file,
                stdout=subprocess.PIPE,
            )
            stdout, stderr = proc.communicate()
            num = stdout.decode("utf8")
            matches = re.search(r"gaussians (\d+)", num)
            num_gauss = int(matches.groups()[0])
        if os.path.exists(self.model_path):
            os.remove(init_log_path)
        os.remove(temp_feats_path)
        self.initial_gaussians = num_gauss
        self.current_gaussians = num_gauss
        self.compile_train_graphs()
        self.mono_align_equal()



"""Utility functions for command line commands"""
from __future__ import annotations

import os

import yaml

from ..exceptions import (
    FileArgumentNotFoundError,
    ModelExtensionError,
    ModelTypeNotSupportedError,
    NoDefaultSpeakerDictionaryError,
    PretrainedModelNotFoundError,
)
from ..models import MODEL_TYPES

__all__ = ["validate_model_arg"]


def validate_model_arg(name: str, model_type: str) -> str:
    """
    Validate pretrained model name argument

    Parameters
    ----------
    name: str
        Name of model
    model_type: str
        Type of model

    Returns
    -------
    str
        Full path of validated model

    Raises
    ------
    :class:`~montreal_forced_aligner.exceptions.ModelTypeNotSupportedError`
        If the type of model is not supported
    :class:`~montreal_forced_aligner.exceptions.FileArgumentNotFoundError`
        If the file specified is not found
    :class:`~montreal_forced_aligner.exceptions.PretrainedModelNotFoundError`
        If the pretrained model specified is not found
    :class:`~montreal_forced_aligner.exceptions.ModelExtensionError`
        If the extension is not valid for the specified model type
    :class:`~montreal_forced_aligner.exceptions.NoDefaultSpeakerDictionaryError`
        If a multispeaker dictionary does not have a default dictionary
    """
    if model_type not in MODEL_TYPES:
        raise ModelTypeNotSupportedError(model_type, MODEL_TYPES)
    model_class = MODEL_TYPES[model_type]
    available_models = model_class.get_available_models()
    model_class = MODEL_TYPES[model_type]
    if name in available_models:
        name = model_class.get_pretrained_path(name)
    elif model_class.valid_extension(name):
        if not os.path.exists(name):
            raise FileArgumentNotFoundError(name)
        if model_type == "dictionary" and os.path.splitext(name)[1].lower() == ".yaml":
            with open(name, "r", encoding="utf8") as f:
                data = yaml.safe_load(f)
                found_default = False
                for speaker, path in data.items():
                    if speaker == "default":
                        found_default = True
                    path = validate_model_arg(path, "dictionary")
                if not found_default:
                    raise NoDefaultSpeakerDictionaryError()
    else:
        if os.path.splitext(name)[1]:
            raise ModelExtensionError(name, model_type, model_class.extensions)
        else:
            raise PretrainedModelNotFoundError(name, model_type, available_models)
    return name



"""
Transcription functions
-----------------------

"""
from __future__ import annotations

import os
import re
import subprocess
import sys
from typing import TYPE_CHECKING, NamedTuple, TextIO

from ..abc import MetaDict
from ..utils import thirdparty_binary

if TYPE_CHECKING:
    from ..abc import MappingType


__all__ = [
    "compose_g",
    "compose_lg",
    "compose_clg",
    "compose_hclg",
    "compose_g_carpa",
    "fmllr_rescore_func",
    "final_fmllr_est_func",
    "initial_fmllr_func",
    "lat_gen_fmllr_func",
    "score_func",
    "lm_rescore_func",
    "carpa_lm_rescore_func",
    "decode_func",
    "create_hclg_func",
]


class CreateHclgArguments(NamedTuple):
    """Arguments for :func:`~montreal_forced_aligner.transcription.multiprocessing.create_hclg_func`"""

    log_path: str
    working_directory: str
    path_template: str
    words_path: str
    carpa_path: str
    small_arpa_path: str
    medium_arpa_path: str
    big_arpa_path: str
    model_path: str
    disambig_L_path: str
    disambig_int_path: str
    hclg_options: MetaDict
    words_mapping: MappingType

    @property
    def hclg_path(self) -> str:
        return self.path_template.format(file_name="HCLG")


class DecodeArguments(NamedTuple):
    """Arguments for :func:`~montreal_forced_aligner.transcription.multiprocessing.decode_func`"""

    log_path: str
    dictionaries: list[str]
    feature_strings: dict[str, str]
    decode_options: MetaDict
    model_path: str
    lat_paths: dict[str, str]
    words_paths: dict[str, str]
    hclg_paths: dict[str, str]


class ScoreArguments(NamedTuple):
    """Arguments for :func:`~montreal_forced_aligner.transcription.multiprocessing.score_func`"""

    log_path: str
    dictionaries: list[str]
    score_options: MetaDict
    lat_paths: dict[str, str]
    rescored_lat_paths: dict[str, str]
    carpa_rescored_lat_paths: dict[str, str]
    words_paths: dict[str, str]
    tra_paths: dict[str, str]


class LmRescoreArguments(NamedTuple):
    """Arguments for :func:`~montreal_forced_aligner.transcription.multiprocessing.lm_rescore_func`"""

    log_path: str
    dictionaries: list[str]
    lm_rescore_options: MetaDict
    lat_paths: dict[str, str]
    rescored_lat_paths: dict[str, str]
    old_g_paths: dict[str, str]
    new_g_paths: dict[str, str]


class CarpaLmRescoreArguments(NamedTuple):
    """Arguments for :func:`~montreal_forced_aligner.transcription.multiprocessing.carpa_lm_rescore_func`"""

    log_path: str
    dictionaries: list[str]
    lat_paths: dict[str, str]
    rescored_lat_paths: dict[str, str]
    old_g_paths: dict[str, str]
    new_g_paths: dict[str, str]


class InitialFmllrArguments(NamedTuple):
    """Arguments for :func:`~montreal_forced_aligner.transcription.multiprocessing.initial_fmllr_func`"""

    log_path: str
    dictionaries: list[str]
    feature_strings: dict[str, str]
    model_path: str
    fmllr_options: MetaDict
    pre_trans_paths: dict[str, str]
    lat_paths: dict[str, str]
    spk2utt_paths: dict[str, str]


class LatGenFmllrArguments(NamedTuple):
    """Arguments for :func:`~montreal_forced_aligner.transcription.multiprocessing.lat_gen_fmllr_func`"""

    log_path: str
    dictionaries: list[str]
    feature_strings: dict[str, str]
    model_path: str
    decode_options: MetaDict
    words_paths: dict[str, str]
    hclg_paths: dict[str, str]
    tmp_lat_paths: dict[str, str]


class FinalFmllrArguments(NamedTuple):
    """Arguments for :func:`~montreal_forced_aligner.transcription.multiprocessing.final_fmllr_est_func`"""

    log_path: str
    dictionaries: list[str]
    feature_strings: dict[str, str]
    model_path: str
    fmllr_options: MetaDict
    trans_paths: dict[str, str]
    spk2utt_paths: dict[str, str]
    tmp_lat_paths: dict[str, str]


class FmllrRescoreArguments(NamedTuple):
    """Arguments for :func:`~montreal_forced_aligner.transcription.multiprocessing.fmllr_rescore_func`"""

    log_path: str
    dictionaries: list[str]
    feature_strings: dict[str, str]
    model_path: str
    fmllr_options: MetaDict
    tmp_lat_paths: dict[str, str]
    final_lat_paths: dict[str, str]


def compose_lg(dictionary_path: str, small_g_path: str, lg_path: str, log_file: TextIO) -> None:
    """
    Compose an LG.fst

    See Also
    --------
    :kaldi_src:`fsttablecompose`
        Relevant Kaldi binary
    :kaldi_src:`fstdeterminizestar`
        Relevant Kaldi binary
    :kaldi_src:`fstminimizeencoded`
        Relevant Kaldi binary
    :kaldi_src:`fstpushspecial`
        Relevant Kaldi binary


    Parameters
    ----------
    dictionary_path: str
        Path to a lexicon fst file
    small_g_path: str
        Path to the small language model's G.fst
    lg_path: str
        Output path to LG.fst
    log_file: TextIO
        Log file handler to output logging info to
    """
    if os.path.exists(lg_path):
        return
    temp_compose_path = lg_path + ".temp"
    compose_proc = subprocess.Popen(
        [thirdparty_binary("fsttablecompose"), dictionary_path, small_g_path, temp_compose_path],
        stderr=log_file,
        env=os.environ,
    )
    compose_proc.communicate()

    temp2_compose_path = lg_path + ".temp2"
    determinize_proc = subprocess.Popen(
        [
            thirdparty_binary("fstdeterminizestar"),
            "--use-log=true",
            temp_compose_path,
            temp2_compose_path,
        ],
        stderr=log_file,
        env=os.environ,
    )
    determinize_proc.communicate()
    os.remove(temp_compose_path)

    minimize_proc = subprocess.Popen(
        [thirdparty_binary("fstminimizeencoded"), temp2_compose_path, temp_compose_path],
        stdout=subprocess.PIPE,
        stderr=log_file,
        env=os.environ,
    )
    minimize_proc.communicate()
    os.remove(temp2_compose_path)
    push_proc = subprocess.Popen(
        [thirdparty_binary("fstpushspecial"), temp_compose_path, lg_path],
        stderr=log_file,
        env=os.environ,
    )
    push_proc.communicate()
    os.remove(temp_compose_path)


def compose_clg(
    in_disambig: str,
    out_disambig: str,
    context_width: int,
    central_pos: int,
    ilabels_temp: str,
    lg_path: str,
    clg_path: str,
    log_file: TextIO,
) -> None:
    """
    Compose a CLG.fst

    See Also
    --------
    :kaldi_src:`fstcomposecontext`
        Relevant Kaldi binary
    :openfst_src:`fstarcsort`
        Relevant OpenFst binary

    Parameters
    ----------
    in_disambig: str
        Path to read disambiguation symbols file
    out_disambig: str
        Path to write disambiguation symbols file
    context_width: int
        Context width of the acoustic model
    central_pos: int
        Central position of the acoustic model
    ilabels_temp:
        Temporary file for ilabels
    lg_path: str
        Path to a LG.fst file
    clg_path:
        Path to save CLG.fst file
    log_file: TextIO
        Log file handler to output logging info to
    """
    compose_proc = subprocess.Popen(
        [
            thirdparty_binary("fstcomposecontext"),
            f"--context-size={context_width}",
            f"--central-position={central_pos}",
            f"--read-disambig-syms={in_disambig}",
            f"--write-disambig-syms={out_disambig}",
            ilabels_temp,
            lg_path,
        ],
        stdout=subprocess.PIPE,
        stderr=log_file,
    )
    sort_proc = subprocess.Popen(
        [thirdparty_binary("fstarcsort"), "--sort_type=ilabel", "-", clg_path],
        stdin=compose_proc.stdout,
        stderr=log_file,
        env=os.environ,
    )
    sort_proc.communicate()


def compose_hclg(
    model_directory: str,
    ilabels_temp: str,
    transition_scale: float,
    clg_path: str,
    hclga_path: str,
    log_file: TextIO,
) -> None:
    """
    Compost HCLG.fst for a dictionary

    See Also
    --------
    :kaldi_src:`make-h-transducer`
        Relevant Kaldi binary
    :kaldi_src:`fsttablecompose`
        Relevant Kaldi binary
    :kaldi_src:`fstdeterminizestar`
        Relevant Kaldi binary
    :kaldi_src:`fstrmsymbols`
        Relevant Kaldi binary
    :kaldi_src:`fstrmepslocal`
        Relevant Kaldi binary
    :kaldi_src:`fstminimizeencoded`
        Relevant Kaldi binary
    :openfst_src:`fstarcsort`
        Relevant OpenFst binary

    Parameters
    ----------
    model_directory: str
        Model working directory with acoustic model information
    ilabels_temp: str
        Path to temporary ilabels file
    transition_scale: float
        Transition scale for the fst
    clg_path: str
        Path to CLG.fst file
    hclga_path: str
        Path to save HCLGa.fst file
    log_file: TextIO
        Log file handler to output logging info to
    """
    model_path = os.path.join(model_directory, "final.mdl")
    tree_path = os.path.join(model_directory, "tree")
    ha_path = hclga_path.replace("HCLGa", "Ha")
    ha_out_disambig = os.path.join(model_directory, "disambig_tid.int")
    make_h_proc = subprocess.Popen(
        [
            thirdparty_binary("make-h-transducer"),
            f"--disambig-syms-out={ha_out_disambig}",
            f"--transition-scale={transition_scale}",
            ilabels_temp,
            tree_path,
            model_path,
            ha_path,
        ],
        stderr=log_file,
        stdout=log_file,
        env=os.environ,
    )
    make_h_proc.communicate()

    temp_compose_path = hclga_path + ".temp"
    compose_proc = subprocess.Popen(
        [thirdparty_binary("fsttablecompose"), ha_path, clg_path, temp_compose_path],
        stderr=log_file,
        env=os.environ,
    )
    compose_proc.communicate()

    determinize_proc = subprocess.Popen(
        [thirdparty_binary("fstdeterminizestar"), "--use-log=true", temp_compose_path],
        stdout=subprocess.PIPE,
        stderr=log_file,
        env=os.environ,
    )
    rmsymbols_proc = subprocess.Popen(
        [thirdparty_binary("fstrmsymbols"), ha_out_disambig],
        stdin=determinize_proc.stdout,
        stdout=subprocess.PIPE,
        stderr=log_file,
        env=os.environ,
    )
    rmeps_proc = subprocess.Popen(
        [thirdparty_binary("fstrmepslocal")],
        stdin=rmsymbols_proc.stdout,
        stdout=subprocess.PIPE,
        stderr=log_file,
        env=os.environ,
    )
    minimize_proc = subprocess.Popen(
        [thirdparty_binary("fstminimizeencoded"), "-", hclga_path],
        stdin=rmeps_proc.stdout,
        stderr=log_file,
        env=os.environ,
    )
    minimize_proc.communicate()
    os.remove(temp_compose_path)


def compose_g(arpa_path: str, words_path: str, g_path: str, log_file: TextIO) -> None:
    """
    Create G.fst from an ARPA formatted language model

    See Also
    --------
    :kaldi_src:`arpa2fst`
        Relevant Kaldi binary

    Parameters
    ----------
    arpa_path: str
        Path to ARPA file
    words_path: str
        Path to words symbols file
    g_path: str
        Path to output G.fst file
    log_file: TextIO
        Log file handler to output logging info to
    """
    arpafst_proc = subprocess.Popen(
        [
            thirdparty_binary("arpa2fst"),
            "--disambig-symbol=#0",
            f"--read-symbol-table={words_path}",
            arpa_path,
            g_path,
        ],
        stderr=log_file,
        stdout=log_file,
    )
    arpafst_proc.communicate()


def compose_g_carpa(
    in_carpa_path: str,
    temp_carpa_path: str,
    words_mapping: MappingType,
    carpa_path: str,
    log_file: TextIO,
):
    """
    Compose a large ARPA model into a G.carpa file

    See Also
    --------
    :kaldi_src:`arpa-to-const-arpa`
        Relevant Kaldi binary

    Parameters
    ----------
    in_carpa_path: str
        Input ARPA model path
    temp_carpa_path: str
        Temporary CARPA model path
    words_mapping: dict[str, int]
        Words symbols mapping
    carpa_path: str
        Path to save output G.carpa
    log_file: TextIO
        Log file handler to output logging info to
    """
    bos_symbol = words_mapping["<s>"]
    eos_symbol = words_mapping["</s>"]
    unk_symbol = words_mapping["<unk>"]
    with open(in_carpa_path, "r", encoding="utf8") as f, open(
        temp_carpa_path, "w", encoding="utf8"
    ) as outf:
        current_order = -1
        num_oov_lines = 0
        for line in f:
            line = line.strip()
            col = line.split()
            if current_order == -1 and not re.match(r"^\\data\\$", line):
                continue
            if re.match(r"^\\data\\$", line):
                log_file.write(r"Processing data...\n")
                current_order = 0
                outf.write(line + "\n")
            elif re.match(r"^\\[0-9]*-grams:$", line):
                current_order = int(re.sub(r"\\([0-9]*)-grams:$", r"\1", line))
                log_file.write(f"Processing {current_order} grams...\n")
                outf.write(line + "\n")
            elif re.match(r"^\\end\\$", line):
                outf.write(line + "\n")
            elif not line:
                if current_order >= 1:
                    outf.write("\n")
            else:
                if current_order == 0:
                    outf.write(line + "\n")
                else:
                    if len(col) > 2 + current_order or len(col) < 1 + current_order:
                        raise Exception(f'Bad line in arpa lm "{line}"')
                    prob = col.pop(0)
                    is_oov = False
                    for i in range(current_order):
                        try:
                            col[i] = str(words_mapping[col[i]])
                        except KeyError:
                            is_oov = True
                            num_oov_lines += 1
                            break
                    if not is_oov:
                        rest_of_line = " ".join(col)
                        outf.write(f"{prob}\t{rest_of_line}\n")
    carpa_proc = subprocess.Popen(
        [
            thirdparty_binary("arpa-to-const-arpa"),
            f"--bos-symbol={bos_symbol}",
            f"--eos-symbol={eos_symbol}",
            f"--unk-symbol={unk_symbol}",
            temp_carpa_path,
            carpa_path,
        ],
        stdin=subprocess.PIPE,
        stderr=log_file,
        stdout=log_file,
        env=os.environ,
    )
    carpa_proc.communicate()
    os.remove(temp_carpa_path)


def create_hclg_func(
    log_path: str,
    working_directory: str,
    path_template: str,
    words_path: str,
    carpa_path: str,
    small_arpa_path: str,
    medium_arpa_path: str,
    big_arpa_path: str,
    model_path: str,
    disambig_L_path: str,
    disambig_int_path: str,
    hclg_options: MetaDict,
    words_mapping: MappingType,
):
    """
    Create HCLG.fst file

    See Also
    --------
    :meth:`.Transcriber.create_hclgs`
        Main function that calls this function in parallel
    :meth:`.Transcriber.create_hclgs_arguments`
        Job method for generating arguments for this function
    :kaldi_src:`add-self-loops`
        Relevant Kaldi binary
    :openfst_src:`fstconvert`
        Relevant OpenFst binary

    Parameters
    ----------
    log_path: str
        Path to log all stderr
    working_directory: str
        Current working directory
    path_template: str
        Path template to construct FST names from
    words_path: str
        Path to word symbols file
    carpa_path: str
        Path to G.carpa file
    small_arpa_path: str
        Path to small ARPA file
    medium_arpa_path: str
        Path to medium ARPA file
    big_arpa_path: str
        Path to big ARPA file
    model_path: str
        Path to acoustic model file
    disambig_L_path: str
        Path to L_disambig.fst file
    disambig_int_path: str
        Path to dictionary's disambiguation symbols file
    hclg_options: dict[str, Any]
        Configuration options for composing HCLG.fst
    words_mapping: dict[str, int]
        Word labels to integer ID mapping
    """
    hclg_path = path_template.format(file_name="HCLG")
    small_g_path = path_template.format(file_name="G.small")
    medium_g_path = path_template.format(file_name="G.med")
    lg_path = path_template.format(file_name="LG")
    hclga_path = path_template.format(file_name="HCLGa")
    if os.path.exists(hclg_path):
        return
    with open(log_path, "w") as log_file:
        context_width = hclg_options["context_width"]
        central_pos = hclg_options["central_pos"]

        clg_path = path_template.format(file_name=f"CLG_{context_width}_{central_pos}")
        ilabels_temp = path_template.format(
            file_name=f"ilabels_{context_width}_{central_pos}"
        ).replace(".fst", "")
        out_disambig = path_template.format(
            file_name=f"disambig_ilabels_{context_width}_{central_pos}"
        ).replace(".fst", ".int")

        log_file.write("Generating decoding graph...\n")
        if not os.path.exists(small_g_path):
            log_file.write("Generating small_G.fst...")
            compose_g(small_arpa_path, words_path, small_g_path, log_file)
        if not os.path.exists(medium_g_path):
            log_file.write("Generating med_G.fst...")
            compose_g(medium_arpa_path, words_path, medium_g_path, log_file)
        if not os.path.exists(carpa_path):
            log_file.write("Generating G.carpa...")
            temp_carpa_path = carpa_path + ".temp"
            compose_g_carpa(big_arpa_path, temp_carpa_path, words_mapping, carpa_path, log_file)
        if not os.path.exists(lg_path):
            log_file.write("Generating LG.fst...")
            compose_lg(disambig_L_path, small_g_path, lg_path, log_file)
        if not os.path.exists(clg_path):
            log_file.write("Generating CLG.fst...")
            compose_clg(
                disambig_int_path,
                out_disambig,
                context_width,
                central_pos,
                ilabels_temp,
                lg_path,
                clg_path,
                log_file,
            )
        if not os.path.exists(hclga_path):
            log_file.write("Generating HCLGa.fst...")
            compose_hclg(
                working_directory,
                ilabels_temp,
                hclg_options["transition_scale"],
                clg_path,
                hclga_path,
                log_file,
            )
        log_file.write("Generating HCLG.fst...")
        self_loop_proc = subprocess.Popen(
            [
                thirdparty_binary("add-self-loops"),
                f"--self-loop-scale={hclg_options['self_loop_scale']}",
                "--reorder=true",
                model_path,
                hclga_path,
            ],
            stderr=log_file,
            stdout=subprocess.PIPE,
            env=os.environ,
        )
        convert_proc = subprocess.Popen(
            [
                thirdparty_binary("fstconvert"),
                "--v=100",
                "--fst_type=const",
                "-",
                hclg_path,
            ],
            stdin=self_loop_proc.stdout,
            stderr=log_file,
            env=os.environ,
        )
        convert_proc.communicate()
        if os.path.exists(hclg_path):
            log_file.write(f"Done generating {hclg_path}!")
        else:
            log_file.write(f"There was an error in generating {hclg_path}")


def decode_func(
    log_path: str,
    dictionaries: list[str],
    feature_strings: dict[str, str],
    decode_options: MetaDict,
    model_path: str,
    lat_paths: dict[str, str],
    word_symbol_paths: dict[str, str],
    hclg_paths: dict[str, str],
) -> None:
    """
    Multiprocessing function for performing decoding

    See Also
    --------
    :meth:`.Transcriber.transcribe`
        Main function that calls this function in parallel
    :meth:`.Transcriber.decode_arguments`
        Job method for generating arguments for this function
    :kaldi_src:`gmm-latgen-faster`
        Relevant Kaldi binary

    Parameters
    ----------
    log_path: str
        Path to save log output
    dictionaries: list[str]
        List of dictionary names
    feature_strings: dict[str, str]
        Dictionary of feature strings per dictionary name
    decode_options: dict[str, Any]
        Options for decoding
    model_path: str
        Path to acoustic model file
    lat_paths: dict[str, str]
        Dictionary of lattice archive paths per dictionary name
    word_symbol_paths: dict[str, str]
        Dictionary of word symbol paths per dictionary name
    hclg_paths: dict[str, str]
        Dictionary of HCLG.fst paths per dictionary name
    """
    with open(log_path, "w", encoding="utf8") as log_file:
        for dict_name in dictionaries:
            feature_string = feature_strings[dict_name]
            lat_path = lat_paths[dict_name]
            word_symbol_path = word_symbol_paths[dict_name]
            hclg_path = hclg_paths[dict_name]
            if os.path.exists(lat_path):
                continue
            if (
                decode_options["uses_speaker_adaptation"]
                and decode_options["first_beam"] is not None
            ):
                beam = decode_options["first_beam"]
            else:
                beam = decode_options["beam"]
            if (
                decode_options["uses_speaker_adaptation"]
                and decode_options["first_max_active"] is not None
            ):
                max_active = decode_options["first_max_active"]
            else:
                max_active = decode_options["max_active"]
            decode_proc = subprocess.Popen(
                [
                    thirdparty_binary("gmm-latgen-faster"),
                    f"--max-active={max_active}",
                    f"--beam={beam}",
                    f"--lattice-beam={decode_options['lattice_beam']}",
                    "--allow-partial=true",
                    f"--word-symbol-table={word_symbol_path}",
                    f"--acoustic-scale={decode_options['acoustic_scale']}",
                    model_path,
                    hclg_path,
                    feature_string,
                    f"ark:{lat_path}",
                ],
                stderr=log_file,
                env=os.environ,
            )
            decode_proc.communicate()


def score_func(
    log_path: str,
    dictionaries: list[str],
    score_options: MetaDict,
    lat_paths: dict[str, str],
    rescored_lat_paths: dict[str, str],
    carpa_rescored_lat_paths: dict[str, str],
    words_paths: dict[str, str],
    tra_paths: dict[str, str],
) -> None:
    """
    Multiprocessing function for scoring lattices

    See Also
    --------
    :func:`~montreal_forced_aligner.transcription.Transcriber.score_transcriptions`
        Main function that calls this function in parallel
    :meth:`.Transcriber.score_arguments`
        Job method for generating arguments for this function
    :kaldi_src:`lattice-scale`
        Relevant Kaldi binary
    :kaldi_src:`lattice-add-penalty`
        Relevant Kaldi binary
    :kaldi_src:`lattice-best-path`
        Relevant Kaldi binary

    Parameters
    ----------
    log_path: str
        Path to save log output
    dictionaries: list[str]
        List of dictionary names
    score_options: dict[str, Any]
        Options for scoring
    lat_paths: dict[str, str]
        Dictionary of lattice archive paths per dictionary name
    rescored_lat_paths: dict[str, str]
        Dictionary of medium G.fst rescored lattice archive paths per dictionary name
    carpa_rescored_lat_paths: dict[str, str]
        Dictionary of carpa-rescored lattice archive paths per dictionary name
    words_paths: dict[str, str]
        Dictionary of word symbol paths per dictionary name
    tra_paths: dict[str, str]
        Dictionary of transcription archive paths per dictionary name
    """
    with open(log_path, "w", encoding="utf8") as log_file:
        for dict_name in dictionaries:
            language_model_weight = score_options["language_model_weight"]
            word_insertion_penalty = score_options["word_insertion_penalty"]
            carpa_rescored_lat_path = carpa_rescored_lat_paths[dict_name]
            rescored_lat_path = rescored_lat_paths[dict_name]
            lat_path = lat_paths[dict_name]
            words_path = words_paths[dict_name]
            tra_path = tra_paths[dict_name]
            if os.path.exists(carpa_rescored_lat_path):
                lat_path = carpa_rescored_lat_path
            elif os.path.exists(rescored_lat_path):
                lat_path = rescored_lat_path
            scale_proc = subprocess.Popen(
                [
                    thirdparty_binary("lattice-scale"),
                    f"--inv-acoustic-scale={language_model_weight}",
                    f"ark:{lat_path}",
                    "ark:-",
                ],
                stdout=subprocess.PIPE,
                stderr=log_file,
                env=os.environ,
            )
            penalty_proc = subprocess.Popen(
                [
                    thirdparty_binary("lattice-add-penalty"),
                    f"--word-ins-penalty={word_insertion_penalty}",
                    "ark:-",
                    "ark:-",
                ],
                stdin=scale_proc.stdout,
                stdout=subprocess.PIPE,
                stderr=log_file,
                env=os.environ,
            )
            best_path_proc = subprocess.Popen(
                [
                    thirdparty_binary("lattice-best-path"),
                    f"--word-symbol-table={words_path}",
                    "ark:-",
                    f"ark,t:{tra_path}",
                ],
                stdin=penalty_proc.stdout,
                stderr=log_file,
                env=os.environ,
            )
            best_path_proc.communicate()


def lm_rescore_func(
    log_path: str,
    dictionaries: list[str],
    lm_rescore_options: MetaDict,
    lat_paths: dict[str, str],
    rescored_lat_paths: dict[str, str],
    old_g_paths: dict[str, str],
    new_g_paths: dict[str, str],
) -> None:
    """
    Multiprocessing function rescore lattices by replacing the small G.fst with the medium G.fst

    See Also
    --------
    :func:`~montreal_forced_aligner.transcription.Transcriber.transcribe`
        Main function that calls this function in parallel
    :meth:`.Transcriber.lm_rescore_arguments`
        Job method for generating arguments for this function
    :kaldi_src:`lattice-lmrescore-pruned`
        Relevant Kaldi binary
    :openfst_src:`fstproject`
        Relevant OpenFst binary

    Parameters
    ----------
    log_path: str
        Path to save log output
    dictionaries: list[str]
        List of dictionary names
    lm_rescore_options: dict[str, Any]
        Options for rescoring
    lat_paths: dict[str, str]
        Dictionary of lattice archive paths per dictionary name
    rescored_lat_paths: dict[str, str]
        Dictionary of rescored lattice archive paths per dictionary name
    old_g_paths: dict[str, str]
        Dictionary of small G.fst paths per dictionary name
    new_g_paths: dict[str, str]
        Dictionary of medium G.fst paths per dictionary name
    """
    with open(log_path, "w", encoding="utf8") as log_file:
        for dict_name in dictionaries:
            lat_path = lat_paths[dict_name]
            rescored_lat_path = rescored_lat_paths[dict_name]
            old_g_path = old_g_paths[dict_name]
            new_g_path = new_g_paths[dict_name]
            if sys.platform == "win32":
                project_type_arg = "--project_output=true"
            else:
                project_type_arg = "--project_type=output"
            if os.path.exists(rescored_lat_path):
                continue

            project_proc = subprocess.Popen(
                [thirdparty_binary("fstproject"), project_type_arg, old_g_path],
                stdout=subprocess.PIPE,
                stderr=log_file,
                env=os.environ,
            )
            lattice_scale_proc = subprocess.Popen(
                [
                    thirdparty_binary("lattice-lmrescore-pruned"),
                    f"--acoustic-scale={lm_rescore_options['acoustic_scale']}",
                    "-",
                    f"fstproject {project_type_arg} {new_g_path} |",
                    f"ark:{lat_path}",
                    f"ark:{rescored_lat_path}",
                ],
                stdin=project_proc.stdout,
                stderr=log_file,
                env=os.environ,
            )
            lattice_scale_proc.communicate()


def carpa_lm_rescore_func(
    log_path: str,
    dictionaries: list[str],
    lat_paths: dict[str, str],
    rescored_lat_paths: dict[str, str],
    old_g_paths: dict[str, str],
    new_g_paths: dict[str, str],
) -> None:
    """
    Multiprocessing function to rescore lattices by replacing medium G.fst with large G.carpa

    See Also
    --------
    :func:`~montreal_forced_aligner.transcription.Transcriber.transcribe`
        Main function that calls this function in parallel
    :meth:`.Transcriber.carpa_lm_rescore_arguments`
        Job method for generating arguments for this function
    :openfst_src:`fstproject`
        Relevant OpenFst binary
    :kaldi_src:`lattice-lmrescore`
        Relevant Kaldi binary
    :kaldi_src:`lattice-lmrescore-const-arpa`
        Relevant Kaldi binary

    Parameters
    ----------
    log_path: str
        Path to save log output
    dictionaries: list[str]
        List of dictionary names
    lat_paths: dict[str, str]
        Dictionary of lattice archive paths per dictionary name
    rescored_lat_paths: dict[str, str]
        Dictionary of rescored lattice archive paths per dictionary name
    old_g_paths: dict[str, str]
        Dictionary of medium G.fst paths per dictionary name
    new_g_paths: dict[str, str]
        Dictionary of large G.carpa paths per dictionary name
    """
    with open(log_path, "w", encoding="utf8") as log_file:
        for dict_name in dictionaries:
            if sys.platform == "win32":
                project_type_arg = "--project_output=true"
            else:
                project_type_arg = "--project_type=output"
            lat_path = lat_paths[dict_name]
            rescored_lat_path = rescored_lat_paths[dict_name]
            old_g_path = old_g_paths[dict_name]
            new_g_path = new_g_paths[dict_name]
            if os.path.exists(rescored_lat_path):
                continue
            project_proc = subprocess.Popen(
                [thirdparty_binary("fstproject"), project_type_arg, old_g_path],
                stdout=subprocess.PIPE,
                stderr=log_file,
                env=os.environ,
            )
            lmrescore_proc = subprocess.Popen(
                [
                    thirdparty_binary("lattice-lmrescore"),
                    "--lm-scale=-1.0",
                    f"ark:{lat_path}",
                    "-",
                    "ark:-",
                ],
                stdout=subprocess.PIPE,
                stdin=project_proc.stdout,
                stderr=log_file,
                env=os.environ,
            )
            lmrescore_const_proc = subprocess.Popen(
                [
                    thirdparty_binary("lattice-lmrescore-const-arpa"),
                    "--lm-scale=1.0",
                    "ark:-",
                    new_g_path,
                    f"ark:{rescored_lat_path}",
                ],
                stdin=lmrescore_proc.stdout,
                stderr=log_file,
                env=os.environ,
            )
            lmrescore_const_proc.communicate()


def initial_fmllr_func(
    log_path: str,
    dictionaries: list[str],
    feature_strings: dict[str, str],
    model_path: str,
    fmllr_options: MetaDict,
    trans_paths: dict[str, str],
    lat_paths: dict[str, str],
    spk2utt_paths: dict[str, str],
) -> None:
    """
    Multiprocessing function for running initial fMLLR calculation

    See Also
    --------
    :func:`~montreal_forced_aligner.transcription.Transcriber.transcribe_fmllr`
        Main function that calls this function in parallel
    :meth:`.Transcriber.initial_fmllr_arguments`
        Job method for generating arguments for this function
    :kaldi_src:`lattice-to-post`
        Relevant Kaldi binary
    :kaldi_src:`weight-silence-post`
        Relevant Kaldi binary
    :kaldi_src:`gmm-post-to-gpost`
        Relevant Kaldi binary
    :kaldi_src:`gmm-est-fmllr-gpost`
        Relevant Kaldi binary

    Parameters
    ----------
    log_path: str
        Path to save log output
    dictionaries: list[str]
        List of dictionary names
    feature_strings: dict[str, str]
        Dictionary of feature strings per dictionary name
    model_path: str
        Path to acoustic model file
    fmllr_options: dict[str, Any]
        Options for calculating fMLLR transforms
    trans_paths: dict[str, str]
        Dictionary of transform archives per dictionary name
    lat_paths: dict[str, str]
        Dictionary of lattice archive paths per dictionary name
    spk2utt_paths: dict[str, str]
        Dictionary of spk2utt scp files per dictionary name
    """
    with open(log_path, "w", encoding="utf8") as log_file:
        for dict_name in dictionaries:
            lat_path = lat_paths[dict_name]
            feature_string = feature_strings[dict_name]
            spk2utt_path = spk2utt_paths[dict_name]
            trans_path = trans_paths[dict_name]

            latt_post_proc = subprocess.Popen(
                [
                    thirdparty_binary("lattice-to-post"),
                    f"--acoustic-scale={fmllr_options['acoustic_scale']}",
                    f"ark:{lat_path}",
                    "ark:-",
                ],
                stdout=subprocess.PIPE,
                stderr=log_file,
                env=os.environ,
            )
            weight_silence_proc = subprocess.Popen(
                [
                    thirdparty_binary("weight-silence-post"),
                    f"{fmllr_options['silence_weight']}",
                    fmllr_options["sil_phones"],
                    model_path,
                    "ark:-",
                    "ark:-",
                ],
                stdin=latt_post_proc.stdout,
                stdout=subprocess.PIPE,
                stderr=log_file,
                env=os.environ,
            )
            gmm_gpost_proc = subprocess.Popen(
                [
                    thirdparty_binary("gmm-post-to-gpost"),
                    model_path,
                    feature_string,
                    "ark:-",
                    "ark:-",
                ],
                stdin=weight_silence_proc.stdout,
                stdout=subprocess.PIPE,
                stderr=log_file,
                env=os.environ,
            )
            fmllr_proc = subprocess.Popen(
                [
                    thirdparty_binary("gmm-est-fmllr-gpost"),
                    f"--fmllr-update-type={fmllr_options['fmllr_update_type']}",
                    f"--spk2utt=ark:{spk2utt_path}",
                    model_path,
                    feature_string,
                    "ark,s,cs:-",
                    f"ark:{trans_path}",
                ],
                stdin=gmm_gpost_proc.stdout,
                stdout=subprocess.PIPE,
                stderr=log_file,
                env=os.environ,
            )
            fmllr_proc.communicate()


def lat_gen_fmllr_func(
    log_path: str,
    dictionaries: list[str],
    feature_strings: dict[str, str],
    model_path: str,
    decode_options: MetaDict,
    word_symbol_paths: dict[str, str],
    hclg_paths: dict[str, str],
    tmp_lat_paths: dict[str, str],
) -> None:
    """
    Regenerate lattices using initial fMLLR transforms

    See Also
    --------
    :func:`~montreal_forced_aligner.transcription.Transcriber.transcribe_fmllr`
        Main function that calls this function in parallel
    :meth:`.Transcriber.lat_gen_fmllr_arguments`
        Job method for generating arguments for this function
    :kaldi_src:`gmm-latgen-faster`
        Relevant Kaldi binary

    Parameters
    ----------
    log_path: str
        Path to save log output
    dictionaries: list[str]
        List of dictionary names
    feature_strings: dict[str, str]
        Dictionary of feature strings per dictionary name
    model_path: str
        Path to acoustic model file
    decode_options: dict[str, Any]
        Options for decoding
    word_symbol_paths: dict[str, str]
        Dictionary of word symbol paths per dictionary name
    hclg_paths: dict[str, str]
        Dictionary of HCLG.fst paths per dictionary name
    tmp_lat_paths: dict[str, str]
        Dictionary of temporary lattice archive paths per dictionary name
    """
    with open(log_path, "w", encoding="utf8") as log_file:
        for dict_name in dictionaries:
            feature_string = feature_strings[dict_name]
            words_path = word_symbol_paths[dict_name]
            hclg_path = hclg_paths[dict_name]
            tmp_lat_path = tmp_lat_paths[dict_name]
            lat_gen_proc = subprocess.Popen(
                [
                    thirdparty_binary("gmm-latgen-faster"),
                    f"--max-active={decode_options['max_active']}",
                    f"--beam={decode_options['beam']}",
                    f"--lattice-beam={decode_options['lattice_beam']}",
                    f"--acoustic-scale={decode_options['acoustic_scale']}",
                    "--determinize-lattice=false",
                    "--allow-partial=true",
                    f"--word-symbol-table={words_path}",
                    model_path,
                    hclg_path,
                    feature_string,
                    f"ark:{tmp_lat_path}",
                ],
                stderr=log_file,
                env=os.environ,
            )

            lat_gen_proc.communicate()


def final_fmllr_est_func(
    log_path: str,
    dictionaries: list[str],
    feature_strings: dict[str, str],
    model_path: str,
    fmllr_options: MetaDict,
    trans_paths: dict[str, str],
    spk2utt_paths: dict[str, str],
    tmp_lat_paths: dict[str, str],
) -> None:
    """
    Multiprocessing function for running final fMLLR estimation

    See Also
    --------
    :func:`~montreal_forced_aligner.transcription.Transcriber.transcribe_fmllr`
        Main function that calls this function in parallel
    :meth:`.Transcriber.final_fmllr_arguments`
        Job method for generating arguments for this function
    :kaldi_src:`lattice-determinize-pruned`
        Relevant Kaldi binary
    :kaldi_src:`lattice-to-post`
        Relevant Kaldi binary
    :kaldi_src:`weight-silence-post`
        Relevant Kaldi binary
    :kaldi_src:`gmm-est-fmllr`
        Relevant Kaldi binary
    :kaldi_src:`compose-transforms`
        Relevant Kaldi binary

    Parameters
    ----------
    log_path: str
        Path to save log output
    dictionaries: list[str]
        List of dictionary names
    feature_strings: dict[str, str]
        Dictionary of feature strings per dictionary name
    model_path: str
        Path to acoustic model file
    fmllr_options: dict[str, Any]
        Options for calculating fMLLR transforms
    trans_paths: dict[str, str]
        Dictionary of transform archives per dictionary name
    spk2utt_paths: dict[str, str]
        Dictionary of spk2utt scp files per dictionary name
    tmp_lat_paths: dict[str, str]
        Dictionary of temporary lattice archive paths per dictionary name
    """
    with open(log_path, "w", encoding="utf8") as log_file:
        for dict_name in dictionaries:
            feature_string = feature_strings[dict_name]
            trans_path = trans_paths[dict_name]
            temp_trans_path = trans_path + ".temp"
            spk2utt_path = spk2utt_paths[dict_name]
            tmp_lat_path = tmp_lat_paths[dict_name]
            determinize_proc = subprocess.Popen(
                [
                    thirdparty_binary("lattice-determinize-pruned"),
                    f"--acoustic-scale={fmllr_options['acoustic_scale']}",
                    "--beam=4.0",
                    f"ark:{tmp_lat_path}",
                    "ark:-",
                ],
                stderr=log_file,
                stdout=subprocess.PIPE,
                env=os.environ,
            )

            latt_post_proc = subprocess.Popen(
                [
                    thirdparty_binary("lattice-to-post"),
                    f"--acoustic-scale={fmllr_options['acoustic_scale']}",
                    "ark:-",
                    "ark:-",
                ],
                stdin=determinize_proc.stdout,
                stdout=subprocess.PIPE,
                stderr=log_file,
                env=os.environ,
            )
            weight_silence_proc = subprocess.Popen(
                [
                    thirdparty_binary("weight-silence-post"),
                    f"{fmllr_options['silence_weight']}",
                    fmllr_options["sil_phones"],
                    model_path,
                    "ark:-",
                    "ark:-",
                ],
                stdin=latt_post_proc.stdout,
                stdout=subprocess.PIPE,
                stderr=log_file,
                env=os.environ,
            )
            fmllr_proc = subprocess.Popen(
                [
                    thirdparty_binary("gmm-est-fmllr"),
                    f"--fmllr-update-type={fmllr_options['fmllr_update_type']}",
                    f"--spk2utt=ark:{spk2utt_path}",
                    model_path,
                    feature_string,
                    "ark,s,cs:-",
                    "ark:-",
                ],
                stdin=weight_silence_proc.stdout,
                stdout=subprocess.PIPE,
                stderr=log_file,
                env=os.environ,
            )

            compose_proc = subprocess.Popen(
                [
                    thirdparty_binary("compose-transforms"),
                    "--b-is-affine=true",
                    "ark:-",
                    f"ark:{trans_path}",
                    f"ark:{temp_trans_path}",
                ],
                stderr=log_file,
                stdin=fmllr_proc.stdout,
                env=os.environ,
            )
            compose_proc.communicate()
            os.remove(trans_path)
            os.rename(temp_trans_path, trans_path)


def fmllr_rescore_func(
    log_path: str,
    dictionaries: list[str],
    feature_strings: dict[str, str],
    model_path: str,
    fmllr_options: MetaDict,
    tmp_lat_paths: dict[str, str],
    final_lat_paths: dict[str, str],
) -> None:
    """
    Multiprocessing function to rescore lattices following fMLLR estimation

    See Also
    --------
    :func:`~montreal_forced_aligner.transcription.Transcriber.transcribe_fmllr`
        Main function that calls this function in parallel
    :meth:`.Transcriber.fmllr_rescore_arguments`
        Job method for generating arguments for this function
    :kaldi_src:`gmm-rescore-lattice`
        Relevant Kaldi binary
    :kaldi_src:`lattice-determinize-pruned`
        Relevant Kaldi binary

    Parameters
    ----------
    log_path: str
        Path to save log output
    dictionaries: list[str]
        List of dictionary names
    feature_strings: dict[str, str]
        Dictionary of feature strings per dictionary name
    model_path: str
        Path to acoustic model file
    fmllr_options: dict[str, Any]
        Options for calculating fMLLR transforms
    tmp_lat_paths: dict[str, str]
        Dictionary of temporary lattice archive paths per dictionary name
    final_lat_paths: dict[str, str]
        Dictionary of lattice archive paths per dictionary name
    """
    with open(log_path, "w", encoding="utf8") as log_file:
        for dict_name in dictionaries:
            feature_string = feature_strings[dict_name]
            tmp_lat_path = tmp_lat_paths[dict_name]
            final_lat_path = final_lat_paths[dict_name]
            rescore_proc = subprocess.Popen(
                [
                    thirdparty_binary("gmm-rescore-lattice"),
                    model_path,
                    f"ark:{tmp_lat_path}",
                    feature_string,
                    "ark:-",
                ],
                stdout=subprocess.PIPE,
                stderr=log_file,
                env=os.environ,
            )
            determinize_proc = subprocess.Popen(
                [
                    thirdparty_binary("lattice-determinize-pruned"),
                    f"--acoustic-scale={fmllr_options['acoustic_scale']}",
                    f"--beam={fmllr_options['lattice_beam']}",
                    "ark:-",
                    f"ark:{final_lat_path}",
                ],
                stdin=rescore_proc.stdout,
                stderr=log_file,
                env=os.environ,
            )

            determinize_proc.communicate()
